# Overview

Stumpf's high-level goals include managing workflows for video annotation
applications using workers from both Mechanical Turk and Kitware. Tasks
performed on videos and video segments are the primary means of accomplishing
these goals. Tasks can be defined either as HITs (human intelligence tasks)
or KWITs (Kitware intelligence tasks).

The canonical deployment of Stumpf is located at stumpf.kitware.com and is
available for use by any Kitware employee.

# Vagrant usage

The following will create Stumpf in a virtual machine and serve the api
from locally on port 5000.
```
vagrant up
vagrant ssh
cd /vagrant
flask run --host 0.0.0.0
```

Note: If you run into issues with the database appearing out of sync, you may
have to migrate the database.  Do this by running `flask db upgrade`. If this
doesn't work, note that you can drop the database and reprovision at any time
by running the following commands:
```
psql -U postgres -h localhost -c 'drop database stumpf'  # inside the VM
vagrant provision  # outside the VM
```
See below for more details on migrations and resetting the database.

Test the server by running the following in a new terminal.  You will need to
provide an authorization token, which is generated for you during OAuth.  If
using mocked OAuth, you only need to follow the `Web Client` section below to
allow the backend to generate a new token.  Otherwise, read the OAuth for local
development section.

```sql
-- List the active tokens
SELECT token from token WHERE expires > CURRENT_TIMESTAMP;
```

Use one of the listed tokens to perform a CURL request.

```bash
curl -H "X-Stumpf-Token:<TOKEN>" http://localhost:5000/api/v1/video
```

This will return a json list of videos or `[]` if none have been imported yet.

# Development

## Database Migrations

Database migrations are performed using Alembic. There are many caveats to
writing good migration scripts and it's recommended to read through the
[Alembic documentation](http://alembic.zzzcomputing.com/en/latest/) before
writing a migration, in particular understanding the limitations of Alembic
[auto-generation of migrations](http://alembic.zzzcomputing.com/en/latest/autogenerate.html#what-does-autogenerate-detect-and-what-does-it-not-detect).

Migrations will migrate both the database schema and the data contained
therein. As migrations are only useful for existing databases, if you are
starting without an existing database, see the Resetting the database section
below.

Alembic migration files will have the version number in the file name, and
Alembic will keep the current version of the database in the `version_num`
column of the `alembic_version` table.

The majority of developer usage of Alembic can be distilled into these commands:

* `flask db current` - show the current version of the database from the `alembic_version` table
* `flask db migrate -m "<description of migration>"` - create a migration based on any current changes to the DB models
* `flask db upgrade` - migrate the database fully forward through all migrations (or can migrate to a specific version)
* `flask db downgrade` - migrate the database backward through one migration (or can migrate to a specific version)

Good migrations should support both upgrade and downgrade paths in the case of a rollback scenario. Some downgrade migrations may be destructive, in which case there is no downward migration possible for those destructive parts. Migrations that migrate schemas and data should be kept separate unless the amount of data to migrate is very small, which allows for low downtime deployments. Some seemingly schema-only changes will have impact on data. All migrations should be carefully reviewed for performance impact because this is tricky.

We occasionally [prune database migrations](http://alembic.zzzcomputing.com/en/latest/cookbook.html#building-an-up-to-date-database-from-scratch) that are no longer applicable to live databases.

## Resetting the database

At any time your local development database can be reset by dropping and recreating the database, then having the Stumpf CLI create tables for you. **Warning** This will delete all data in your local `stumpf` database. Using this method, you will not invoke the Alembic migrations, but rather will end up with a database schema that is tied to the models in the current stumpf repo HEAD (and so should be equivalent in theory to running through all of the migrations). It will add an `alembic_version` to the database that is the latest migration version file from the current stumpf repo HEAD.

```
psql -U postgres -h localhost -c "drop database stumpf"
psql -U postgres -h localhost -c "create database stumpf owner stumpf"
stumpf-create-tables
```

Now you will have a versioned database schema but no data. See below for generating fake data.

## Generating fake data

It's often desirable in a development environment to simulate a database full of production-like quality data. In Stumpf there are tools that exist to create any amount of fake data for testing and development. In particular, Stumpf makes use of a tool called faker and wraps fake data generation in a command line script `stumpf-generate-sample-data`. Note that the fake data is not meant to be a replacement for production data but rather a tool for testing things such as database query times, user interface quirks, etc.

Note that any changes to models should also reflect those changes in the fake data provided by stumpf. Ideally at any commit a user would be able to generate fake data for that particular codebase snapshot.

Migrations are primarily intended to target production data and may overlook some aspects of the fake data, so it's possible that a migration will fail because of fake data in your database.

For the impatient, it is possible to generate less fake data with `stumpf-generate-sample-data --small`.

## Web client

For development of the web application, we use webpack-dev-server with a proxy to
api server (defaults to localhost:5000).  You can run this server outside of the
virtual machine.  First, ensure you have a recent version of node installed
(>= 8 recommended).

```
# install yarn globally
npm install -g yarn

# go to the web directory and install dependencies
yarn

# set an environment variable telling the dev server where to proxy
# api requests (defaults to localhost:5000)
export STUMPF_API=http://localhost:5000

# run the development server
yarn dev
```

## Testing

Testing can be run with `tox` from the project root. Tests require postgres, and the following environment variables are supported:

* `SQLALCHEMY_DATABASE_URI` - default: `postgresql://postgres@localhost/stumpf_testdb`

## OAuth for local development

Note: You can skip this section if you just want to use the default mocked
interface for OAuth. 

For authentication to work, you must set environment variables required for oauth
session negotiation.
```
GOOGLE_OAUTH_CLIENT_ID
GOOGLE_OAUTH_CLIENT_SECRET

OAUTHLIB_INSECURE_TRANSPORT=1  # to allow authenticating over plain-text HTTP
```
These values for the deployed application can be found in vault under
`secret/projects/diva-mturk/stumpf/oauth`.  Developers can generate a new oauth
application at  <https://console.developers.google.com/apis/credentials/oauthclient/>
You should set the origin to local host/port created by webpack-dev-server
(<http://localhost:8080> by default).  The redirect path for stumpf is at `/auth/google/authorized`
(e.g. <http://localhost:8080/auth/google/authorized>).
