#!/usr/bin/env python
"""
Target videos with ground truth (from Kellie):

2018-03-15_10-35-00_10-40-00_G299_school:
    This clip is not part of the set of cameras for the deliverable.  This clip
    was selected to give true negative.  I confirmed that no activities occur
    in this clip.
2018-03-15_10-35-00_10-40-00_G328_school:
    This clip is not part of the set of cameras for the  deliverable, but does
    have activities.  I could find a comparable clip that is part of the
    deliverable and has annotations, if you are ok switching it out.
2018-03-15_10-35-00_10-40-00_G329_admin:
    This clip is not part of the set of cameras for the  deliverable.  This
    clip was selected to give few activities  I confirmed that "entering
    through structure" occurs for two people from t = 4:33 - 4:35 and t = 4:36
    - 4:37.
2018-03-15_10-35-00_10-40-00_G330_school:
    This clip is not part of the set of cameras for the  deliverable.  This
    clip was selected to give true negative.  I confirmed that no activities
    occur in this clip.
2018-03-15_10-35-00_10-40-00_G331_bus:
    This clip IS part of the set of cameras for the deliverable.  This data set
    has fully coronated grozper & refiner results.  I have attached the
    relevant files to this email, just let me know if you'd like me to do
    anything further with them.
"""
import json
import os
import time

import requests

# hit options
sandbox = True
delay = 2  # delay between posting tasks for AWS rate limiting
title = 'Determine activities present in short video clips'
description = (
    'Watch a short video clip and answer which of a given list '
    'of activities occur in the clip.  A large screen is recommended '
    'because videos cannot currently be zoomed.  Also, Safari,'
    'Internet Explorer, and Edge are not supported.  We recommend '
    'Firefox or Chrome on a desktop or laptop computer.'
)
max_assignments = 3
duration = 10 * 60
lifetime = 12 * 60 * 60
reward = 10

# hit definition
videos = [
    'e65ea17a-4625-4380-a6a2-fdfa71c6192c',  # 2018-03-15_10-35-00_10-40-00_G299_school
    'b7685395-d85c-46f8-9d12-92bf5a75d785',  # 2018-03-15_10-35-00_10-40-00_G328_school
    '733c0e07-8cc6-4575-9703-51939db02c98',  # 2018-03-15_10-35-00_10-40-00_G329_admin
    '33422e0f-3471-4c58-9f7e-19e7e452cf91',  # 2018-03-15_10-35-00_10-40-00_G330_school
    '9dc79d50-5124-4855-8435-49cfd312bed7'  # 2018-03-15_10-35-00_10-40-00_G331_bus
]

activity_names = {
    'people talking',
    'person entering through structure',
    'person opens trunk',
    'person texting on phone',
    'vehicle turning left'
}

with open('activities.json', 'r') as activity_file:
    activities = {
        key: value for key, value in json.load(activity_file).items()
        if key in activity_names
    }
assert len(activity_names) == len(activities), 'Unknown activity name'

app = '2d755948-af16-40bf-846e-0ce885a43b53'
token = os.environ['STUMPF_TOKEN']


def generate_payload(start, end):
    return {
        'sandbox': sandbox,
        'args': {
            'start': start,
            'end': end,
            'activities': json.dumps(activities)
        },
        'title': title,
        'description': description,
        'max_assignments': max_assignments,
        'assignment_duration_in_seconds': duration,
        'lifetime_in_seconds': lifetime,
        'reward': reward,
        'auto_approval_delay_in_seconds': 10
    }


def generate_hit(video, start, end):
    return (
        f'https://stumpf.kitware.com/api/v1/task/{video}/{app}/hit',
        generate_payload(start, end)
    )


def serialize_hits():
    with open('source_hits.jsonl', 'w') as f:
        for video in videos:
            for clip in range(15):
                start = clip * 20
                end = start + 20
                f.write(json.dumps(generate_hit(video, start, end)) + '\n')


def submit_hit(hit):
    resp = requests.post(hit[0], json=hit[1], headers={'X-Stumpf-Token': token})
    if not resp.ok:
        try:
            print(resp.json())
        except Exception:
            print(resp.text)
    assert resp.ok


def submit_hits():
    with open('source_hits.jsonl', 'r') as source, \
            open('submitted_hits.jsonl', 'w') as finished, \
            open('failed_hits.jsonl', 'w') as failed:

        for line in source.readlines():
            hit = json.loads(line.strip())
            try:
                submit_hit(hit)
                finished.write(line)
            except Exception:
                failed.write(line)
            finished.flush()
            failed.flush()
            time.sleep(delay)


if __name__ == '__main__':
    serialize_hits()
    submit_hits()
