import Raven from 'raven-js';
import axios_ from 'axios';

const devActivities = {
  'Person closes trunk': 'A trunk is defined as a container designed to store non-human cargo on a vehicle (e.g., rear facing trunk, van rear doors, truck bed).',
  'Person enters through structure': 'A person entering (going into the field of view) through a physical structure (e.g., doorway, arch, gateway, subway entrance/exit).',
  'Person exits through structure': 'A person exiting (going out of the field of view) through a physical structure (e.g., doorway, arch, gateway, subway entrance/exit).',
  'Person loads vehicle': 'An object moving from person to vehicle.',
  'Person opens trunk': 'A trunk is defined as a container designed to store non-human cargo on a vehicle (e.g., rear facing trunk, van rear doors, truck bed).',
};
const baseURL = 'https://stumpf.kitware.com/api/mturk/v1';

function captureResponse(resp) {
  // When intercepting an error, the response object is an attribute of the
  // object provided in the argument.
  const response = resp.response || resp;
  try {
    Raven.captureBreadcrumb({
      message: `${resp.config.method.toUpperCase()} ${resp.config.url} => ${response.status}`,
      category: 'axios response',
      level: 'info',
      data: response.data,
    });
  } catch (err) {
    Raven.captureException(err);
  }
  if (resp instanceof Error) {
    return Promise.reject(resp);
  }
  return resp;
}

function captureRequest(config) {
  try {
    Raven.captureBreadcrumb({
      message: `${config.method.toUpperCase()} ${config.url}`,
      category: 'axios request',
      level: 'info',
      data: config.data,
    });
  } catch (err) {
    Raven.captureException(err);
  }
  if (config instanceof Error) {
    return Promise.reject(config);
  }
  return config;
}

class ArityApi { /* eslint-disable no-underscore-dangle */
  constructor(params) {
    const axios = axios_.create({
      baseURL,
    });
    this._stumpfAssignment = 'not authenticated';
    this._videoPath = '';

    axios.interceptors.request.use((config) => {
      if (this.authenticated) {
        return {
          ...config,
          headers: {
            'X-Stumpf-Mturk-Token': this.token,
            ...config.headers,
          },
        };
      }
      return config;
    });

    axios.interceptors.response.use((response) => {
      const { status } = response;
      if (!status || status < 200 || status >= 300) {
        const err = new Error('Bad status');
        Object.assign(err, response);
        return Promise.reject(err);
      }
      return response;
    });

    axios.interceptors.request.use(captureRequest, captureRequest);
    axios.interceptors.response.use(captureResponse, captureResponse);

    Object.assign(this, {
      axios,
      params,
      authenticated: false,
      finished: false,
      submitted: false,
    });
    window.api = this;
  }

  get development() {
    return this.params.development || this.preview;
  }

  get preview() {
    return this.params.assignmentId === 'ASSIGNMENT_ID_NOT_AVAILABLE';
  }

  get assignmentId() {
    return this.params.assignmentId;
  }

  get hitId() {
    return this.params.hitId;
  }

  get workerId() {
    return this.params.workerId;
  }

  get activities() {
    return this.params.activities ? JSON.parse(this.params.activities) : devActivities;
  }

  get valid() {
    return this.params.development || (
      this.assignmentId && this.hitId && this.workerId
    ) || this.preview;
  }

  get start() {
    return parseFloat(this.params.start !== undefined ? this.params.start : 40);
  }

  get end() {
    return parseFloat(this.params.end !== undefined ? this.params.end : 60);
  }

  get stumpfAssignment() {
    return this._stumpfAssignment;
  }

  get videoPath() {
    return this._videoPath;
  }

  get turkSubmitTo() {
    return this.params.turkSubmitTo;
  }

  get context() {
    return {
      preview: this.preview,
      development: this.development,
      assignmentId: this.assignmentId,
      stumpfAssignment: this.stumpfAssignment,
      hitId: this.hitId,
      workerId: this.workerId,
      activities: this.activities,
      valid: this.valid,
      videoPath: this.videoPath,
      start: this.start,
      end: this.end,
    };
  }

  get version() { // eslint-disable-line class-methods-use-this
    return process.env.VUE_APP_GITHASH;
  }

  async login() {
    if (this.development) {
      await new Promise((resolve) => {
        this.authenticated = true;
        setTimeout(resolve, 1);
      });
      this._stumpfAssignment = 'development';
      this._videoPath = 'https://diva-mturk-video-catalog.s3.amazonaws.com/uuid-encoded/a6c670d1-ecaf-4158-b900-38103878f780.webm';
      return {
        assignment: {
          task: {
            video: {
              path: this.videoPath,
            },
            args: {
              activites: this.activities,
              start: this.start,
              end: this.end,
            },
          },
        },
      };
    }

    if (!this.valid) {
      return null;
    }

    const authParams = {
      mturk_assignment_id: this.assignmentId,
      mturk_hit_id: this.hitId,
      mturk_worker_id: this.workerId,
      app_version: this.version,
    };
    Raven.setExtraContext(this.context);

    let resp;
    try {
      resp = await this.axios.post('/authenticate', authParams);
    } catch (err) {
      console.error('Stumpf authentication failed');
      throw err;
    }

    this.token = resp.data.token;
    this._videoPath = resp.data.assignment.task.video.path;
    this._stumpfAssignment = resp.data.assignment.id;
    this.authenticated = true;
    return resp.data;
  }

  async submit(answer) {
    if (this.submitted || this.development || this.preview) {
      this.finished = true;
      console.log(answer);
      return true;
    }

    let resp;
    try {
      await this.axios.post('/answer', { answer });
      this.submitted = true;
    } catch (err) {
      console.error('Posting answer to Stumpf failed');
      throw err;
    }
    return resp;
  }

  async feedback(message) {
    if (this.development || this.preview) {
      console.log(message);
      await new Promise((resolve) => {
        this.finished = true;
        setTimeout(resolve, 1);
      });
      return true;
    }

    let resp;
    try {
      await this.axios.post('/feedback', { args: { feedback: message } });
    } catch (err) {
      console.error('Posting feedback to Stumpf failed');
      throw err;
    }
    return resp;
  }

  async log(data) {
    if (this.development || this.preview || !this.authenticated) {
      console.log(data);
      await new Promise((resolve) => {
        setTimeout(resolve, 1);
      });
      return true;
    }

    let resp;
    try {
      await this.axios.post('/log', { category: 'Arity', message: { data } });
    } catch (err) {
      console.error('Posting log message to Stumpf failed');
      throw err;
    }
    return resp;
  }
}

export default ArityApi;
