import axios_ from 'axios';

const baseURL = 'https://stumpf.kitware.com/api/mturk/v1';

class ArityApi { /* eslint-disable no-underscore-dangle */
  constructor(params) {
    const axios = axios_.create({
      baseURL: params.baseURL ? params.baseURL : baseURL,
    });
    this._stumpfAssignment = 'not authenticated';
    this._videoPath = '';

    axios.interceptors.request.use((config) => {
      if (this.authenticated) {
        return {
          ...config,
          headers: {
            'X-Stumpf-Mturk-Token': this.token,
            'Content-Type': 'application/json',
            ...config.headers,
          },
        };
      }
      return config;
    });

    axios.interceptors.response.use((response) => {
      const { status } = response;
      if (!status || status < 200 || status >= 300) {
        const err = new Error('Bad status');
        Object.assign(err, response);
        return Promise.reject(err);
      }
      return response;
    });

    Object.assign(this, {
      axios,
      params,
      authenticated: false,
      finished: false,
      submitted: false,
    });

    window.api = this;
  }

  get development() {
    return this.params.development || this.preview;
  }

  get preview() {
    return this.params.assignmentId === 'ASSIGNMENT_ID_NOT_AVAILABLE';
  }

  get taskId() {
    return this.params.task_id;
  }

  get valid() {
    return this.params.development || this.taskId || this.preview;
  }

  get start() {
    return parseFloat(this.params.start !== undefined ? this.params.start : 0);
  }

  get end() {
    return parseFloat(this.params.end !== undefined ? this.params.end : 300);
  }

  get framerate() {
    return parseFloat(this._framerate !== undefined ? this._framerate : 30.0);
  }

  get originalVideoWidth() {
    return this._videoWidth;
  }

  get stumpfAssignment() {
    return this._stumpfAssignment;
  }

  get videoPath() {
    return this._videoPath;
  }

  get context() {
    return {
      preview: this.preview,
      development: this.development,
      taskId: this.taskId,
      stumpfAssignment: this.stumpfAssignment,
      valid: this.valid,
      videoPath: this.videoPath,
      start: this.start,
      end: this.end,
    };
  }

  get version() { // eslint-disable-line class-methods-use-this
    return process.env.VUE_APP_GITHASH;
  }

  async login() {
    if (this.development) {
      await new Promise((resolve) => {
        this.authenticated = true;
        setTimeout(resolve, 1);
      });
      this._stumpfAssignment = 'development';
      this._videoPath = 'https://diva-mturk-video-catalog.s3.amazonaws.com/uuid-encoded/a6c670d1-ecaf-4158-b900-38103878f780.webm';
      this._framerate = 30.0;
      this._videoWidth = 1920;
      return {
        assignment: {
          task: {
            video: {
              path: this.videoPath,
            },
            args: {
              activites: this.activities,
              start: this.start,
              end: this.end,
            },
          },
        },
      };
    }

    if (!this.valid) {
      return null;
    }

    const authParams = {
      task_id: this.taskId,
      app_version: this.version,
    };
    let resp;
    try {
      resp = await this.axios.post('/authenticate/kwit', authParams);
    } catch (err) {
      console.error('Stumpf authentication failed');
      throw err;
    }

    this.token = resp.data.token;
    this._videoPath = resp.data.assignment.task.video.path;
    this._stumpfAssignment = resp.data.assignment.id;
    this._framerate = resp.data.assignment.task.video.frame_rate;
    this._videoWidth = resp.data.assignment.task.video.width;
    this.authenticated = true;
    return resp.data;
  }

  async submit(answer, force = false) {
    console.log(answer);
    if (!force && (this.submitted || this.development || this.preview)) {
      this.finished = true;
      return true;
    }
    let resp;
    try {
      await this.axios.post('/answer', { answer });
      this.submitted = true;
      this.finished = true;
    } catch (err) {
      console.error('Posting answer to Stumpf failed');
      throw err;
    }
    return resp;
  }
}

export default ArityApi;
