import Raven from 'raven-js';
import axios_ from 'axios';

const baseURL = 'https://stumpf.kitware.com/api/mturk/v1';

function captureResponse(resp) {
  // When intercepting an error, the response object is an attribute of the
  // object provided in the argument.
  const response = resp.response || resp;
  try {
    Raven.captureBreadcrumb({
      message: `${resp.config.method.toUpperCase()} ${resp.config.url} => ${response.status}`,
      category: 'axios response',
      level: 'info',
      data: response.data,
    });
  } catch (err) {
    Raven.captureException(err);
  }
  if (resp instanceof Error) {
    return Promise.reject(resp);
  }
  return resp;
}

function captureRequest(config) {
  try {
    Raven.captureBreadcrumb({
      message: `${config.method.toUpperCase()} ${config.url}`,
      category: 'axios request',
      level: 'info',
      data: config.data,
    });
  } catch (err) {
    Raven.captureException(err);
  }
  if (config instanceof Error) {
    return Promise.reject(config);
  }
  return config;
}

class StumpfApi { /* eslint-disable no-underscore-dangle */
  constructor(params, taskType = 'hit') {
    const axios = axios_.create({
      baseURL: params.baseURL ? params.baseURL : baseURL,
    });
    this._stumpfAssignment = 'not authenticated';
    this._videoPath = '';

    axios.interceptors.request.use((config) => {
      if (this.authenticated) {
        return {
          ...config,
          headers: {
            'X-Stumpf-Mturk-Token': this.token,
            'Content-Type': 'application/json',
            ...config.headers,
          },
        };
      }
      return config;
    });

    axios.interceptors.response.use((response) => {
      const { status } = response;
      if (!status || status < 200 || status >= 300) {
        const err = new Error('Bad status');
        Object.assign(err, response);
        return Promise.reject(err);
      }
      return response;
    });

    Object.assign(this, {
      axios,
      params,
      taskType,
      authenticated: false,
      finished: false,
      submitted: false,
    });

    if (this.shouldReport) {
      axios.interceptors.request.use(captureRequest, captureRequest);
      axios.interceptors.response.use(captureResponse, captureResponse);
    }
  }

  /* HIT ONLY params -- supplied by mechanical turk */

  get assignmentId() {
    return this.params.assignmentId;
  }

  get hitId() {
    return this.params.hitId;
  }

  get workerId() {
    return this.params.workerId;
  }

  get turkSubmitTo() {
    return this.params.turkSubmitTo;
  }

  /* KWIT only params */

  get taskId() {
    return this.params.taskId;
  }

  /* generic properties for ALL tasks */

  get args() {
    return this._args;
  }

  get preview() {
    return this.params.assignmentId === 'ASSIGNMENT_ID_NOT_AVAILABLE';
  }

  get framerate() {
    return parseFloat(this._framerate !== undefined ? this._framerate : 30.0);
  }

  get originalVideoWidth() {
    return this._videoWidth;
  }

  get originalVideoHeight() {
    return this._videoHeight;
  }

  get stumpfAssignment() {
    return this._stumpfAssignment;
  }

  get videoPath() {
    return this._videoPath;
  }

  /* Derived properties */

  get valid() {
    const validHit = this.assignmentId && this.hitId && this.workerId;
    const validKwit = this.taskId;
    return this.params.development || (
      validHit && this.taskType === 'hit'
    ) || (
        validKwit && this.taskType === 'kwit'
      ) || this.preview;
  }

  get authParams() {
    if (this.taskType === 'hit') {
      return {
        mturk_assignment_id: this.assignmentId,
        mturk_hit_id: this.hitId,
        mturk_worker_id: this.workerId,
        app_version: this.version,
      };
    } else if (this.taskType === 'kwit') {
      return {
        task_id: this.taskId,
        app_version: this.version,
      };
    } else {
      throw new Error(`Invalid task type ${this.taskType}`);
    }
  }

  get shouldReport() {
    return this.taskType === 'hit' && !this.development;
  }

  get development() {
    return this.params.development || this.preview;
  }

  get context() {
    return {
      preview: this.preview,
      development: this.development,
      assignmentId: this.assignmentId,
      stumpfAssignment: this.stumpfAssignment,
      valid: this.valid,
      videoPath: this.videoPath,
    };
  }

  get version() { // eslint-disable-line class-methods-use-this
    return process.env.VUE_APP_GITHASH;
  }

  /* REST Invocations */

  async login(devParams) {
    if (this.development) {
      await new Promise((resolve) => {
        this.authenticated = true;
        setTimeout(resolve, 1);
      });
      this._stumpfAssignment = 'development';
      this._videoPath = devParams.assignment.task.video.path;
      this._videoWidth = devParams.assignment.task.video.width;
      this._videoHeight = devParams.assignment.task.video.height;
      this._args = devParams.assignment.task.args;
      return devParams;
    }

    if (!this.valid) {
      return null;
    }

    Raven.setExtraContext(this.context);

    let resp;
    try {
      const authURI = '/authenticate' + (this.taskType === 'kwit' ? '/kwit' : '');
      resp = await this.axios.post(authURI, this.authParams);
    } catch (err) {
      console.error('Stumpf authentication failed');
      throw err;
    }

    this.token = resp.data.token;
    this._videoPath = resp.data.assignment.task.video.path;
    this._stumpfAssignment = resp.data.assignment.id;
    this._framerate = resp.data.assignment.task.video.frame_rate;
    this._videoWidth = resp.data.assignment.task.video.width;
    this._videoHeight = resp.data.assignment.task.video.height;
    this._args = resp.data.assignment.task.args;
    this.authenticated = true;
    return resp.data;
  }

  async submit(answer) {
    Raven.setExtraContext({ answer });

    if (this.submitted || this.development || this.preview) {
      this.finished = true;
      console.log(answer);
      return true;
    }

    let resp;
    try {
      await this.axios.post('/answer', { answer });
      this.submitted = true;
    } catch (err) {
      console.error('Posting answer to Stumpf failed');
      throw err;
    }
    return resp;
  }

  async feedback(message) {
    if (this.development || this.preview) {
      console.log(message);
      await new Promise((resolve) => {
        this.finished = true;
        setTimeout(resolve, 1);
      });
      return true;
    }

    let resp;
    try {
      await this.axios.post('/feedback', { args: { feedback: message } });
    } catch (err) {
      console.error('Posting feedback to Stumpf failed');
      throw err;
    }
    return resp;
  }

  async log(data) {
    if (this.development || this.preview || !this.authenticated) {
      console.log(data);
      await new Promise((resolve) => {
        setTimeout(resolve, 1);
      });
      return true;
    }

    let resp;
    try {
      await this.axios.post('/log', { category: 'Re-id', message: { data } });
    } catch (err) {
      console.error('Posting log message to Stumpf failed');
      throw err;
    }
    return resp;
  }
}

export default StumpfApi;
