"""
Migrate diva tables to facts

Revision ID: 36af2963f099
Revises: 1ba21a299d27
Create Date: 2018-10-19 08:57:08.491083

"""
from datetime import datetime

from alembic import op
from sqlalchemy.orm.attributes import flag_modified

from stumpf.diva_fact_schemas import ActivityFactPropertySchema, \
                                    TrackDetectionFactPropertySchema, \
                                    TrackFactPropertySchema
from stumpf.diva_models import Activity, ActivitySchema
from stumpf.models import db, Provenance, Task


# revision identifiers, used by Alembic.
revision = '36af2963f099'
down_revision = '1ba21a299d27'
branch_labels = None
depends_on = None


def upgrade():
    activity_schema = ActivityFactPropertySchema()
    detection_schema = TrackDetectionFactPropertySchema()
    track_schema = TrackFactPropertySchema()

    # commit all previous migrations and ensure we are not in autocommit mode
    op.execute('COMMIT')
    connection = op.get_bind()
    connection.execution_options(isolation_level='READ_COMMITTED')

    # migrate diva tables
    activitySchema = ActivitySchema()
    batch = f'migrate-{datetime.now().isoformat()}'
    meta = {
        'batch': batch,
        'migration': True
    }

    for activity_row in Activity.query.all():
        activity_object = activitySchema.dump(activity_row)

        activity_fact, provenance = activity_row.video.get_fact_from_external_process(
            activity_schema.load({
                'type': activity_object['type_']
            }),
            frame_range=activity_object['frame_range'],
            schema='grozper:activity',
            provenance={
                'diva_table': 'activity',
                'diva_id': activity_object['id'],
                **meta
            }
        )
        db.session.add_all([activity_fact, provenance])
        db.session.commit()

        for track_object in activity_object['tracks']:
            track_type = track_object['type_']
            track_fact, provenance = activity_row.video.get_fact_from_external_process(
                track_schema.load({
                    'activity_id': str(activity_fact.id),
                    'type': track_type
                }),
                frame_range=activity_object['frame_range'],
                schema='grozper:track',
                provenance={
                    'diva_table': 'track',
                    'diva_id': track_object['id'],
                    **meta
                }
            )
            db.session.add_all([track_fact, provenance])
            db.session.commit()

            for detection_object in track_object['detections']:
                detection_frame = detection_object['frame']
                bounds = detection_object['bounds']
                detection_fact, provenance = activity_row.video.get_fact_from_external_process(
                    detection_schema.load({
                        'type': track_type,
                        'activity_id': str(activity_fact.id),
                        'track_id': str(track_fact.id),
                        'source': detection_object['source'],
                        'occlusion': detection_object['occlusion'],
                        'left': bounds[0],
                        'right': bounds[2],
                        'top': bounds[1],
                        'bottom': bounds[3]
                    }),
                    frame_range=(detection_frame, detection_frame),
                    schema='grozper:track_detection',
                    provenance={
                        'diva_table': 'detection',
                        'diva_id': detection_object['id'],
                        **meta
                    }
                )
                db.session.add_all([detection_fact, provenance])
            db.session.commit()

    # migrate gladiator tasks
    cursor = Task.query.filter_by(annotation_app_id='d4f8d27a-b3b6-44c3-b6ed-1e8910048c2d')
    for task in cursor.all():
        args = task.args
        if args is None:
            continue

        activity_id = args.get('activityId')
        if activity_id is None:
            continue

        provenance = Provenance.query.filter(
            Provenance.meta['diva_id'].astext == activity_id).first()
        if provenance is None:
            continue

        task.args['activityId'] = str(provenance.fact_id)
        flag_modified(task, 'args')

    db.session.commit()


def downgrade():
    # NOPE, NOPE, NOPE
    pass
