from flask import abort, current_app, g, redirect, request, session
from flask_dance.consumer import oauth_authorized
from flask_dance.consumer.backend.sqla import SQLAlchemyBackend
from flask_dance.consumer.requests import OAuth2Session as BaseOAuth2Session
from flask_dance.contrib.google import make_google_blueprint
from flask_login import current_user, login_user
from flask_login import LoginManager
from sqlalchemy.orm.exc import NoResultFound

from stumpf.decorators import is_public
from stumpf.models import db, MturkToken, OAuth, Token, User


# This is a version of flask-dance's default session class, except that it
# sets the oauth state arg to a value from the flask session.
class OAuth2Session(BaseOAuth2Session):

    def authorization_url(self, url, **kwargs):
        kwargs.pop('state', None)  # ignore the state provided from flask-dance
        state = session.pop('oauth-redirect-destination', '/')
        return super().authorization_url(url, state=state, **kwargs)


auth_bp = make_google_blueprint(backend=SQLAlchemyBackend(OAuth, db.session, user=current_user),
                                reprompt_consent=True, offline=True, scope=['email'],
                                session_class=OAuth2Session)
login_manager = LoginManager()


def before_request_auth(*args, **kwargs):
    # allow unauthenticated OPTIONS requests required for CORS preflight, and
    # allow unauthenticated requests for public endpoints, and endpoints which couldn't be routed
    # (404)
    if is_public() or request.method == 'OPTIONS' or request.routing_exception:
        return
    elif request.blueprint == 'mturk':  # handle mturk API auth
        token = MturkToken.query.filter_by(
            expired=False, token=request.headers.get('X-Stumpf-Mturk-Token')).first()

        if token:
            g.assignment_id = token.assignment_id

            # Add assignment ID to sentry context for debugging
            if hasattr(current_app, 'sentry'):
                current_app.sentry.user_context({
                    'assignment_id': str(g.assignment_id)
                })
        else:
            abort(401)
    else:  # handle generic auth
        user = Token.user_from_token(request.headers.get('X-Stumpf-Token'))

        if user is None:
            login_manager.unauthorized()
        login_user(user)


@login_manager.user_loader
def load_user(user_id):
    return User.query.get(user_id)


# create/login local user on successful OAuth login
def google_logged_in(blueprint, token):
    if not token:
        abort(403)

    resp = blueprint.session.get('/oauth2/v2/userinfo')
    if not resp.ok:
        abort(403)

    google_info = resp.json()

    # hd is only guaranteed to exist for GSuite accounts, see
    # https://developers.google.com/identity/protocols/OpenIDConnect#hd-param
    if google_info.get('hd') != 'kitware.com':
        abort(403)

    oauth_query = OAuth.query.filter_by(
        provider=blueprint.name,
        provider_user_id=google_info['id']
    )

    try:
        oauth = oauth_query.one()
    except NoResultFound:
        oauth = OAuth(
            provider=blueprint.name,
            provider_user_id=google_info['id'],
            token=token
        )

    if oauth.user:
        login_user(oauth.user)
    else:
        try:
            user = User(email=google_info['email'])
            oauth.user = user
            db.session.add_all([user, oauth])
            db.session.commit()
            login_user(user)
        except Exception:
            db.session.rollback()
            raise

    stumpf_token = Token(user_id=current_user.get_id())
    db.session.add(stumpf_token)
    db.session.commit()

    response = redirect(request.args.get('state', '/'))
    response.set_cookie('stumpf-token', stumpf_token.token, expires=stumpf_token.expires)
    return response


@oauth_authorized.connect_via(auth_bp)
def google_logged_in_decorated(*args, **kwargs):
    return google_logged_in(*args, **kwargs)


def mturk_before_request_auth(*args, **kwargs):
    # allow unauthenticated OPTIONS requests required for CORS preflight requests
    if is_public() or request.method == 'OPTIONS':
        return
    else:
        token = MturkToken.query.filter_by(
            expired=False, token=request.headers.get('X-Stumpf-Mturk-Token')).first()

        if token:
            g.assignment_id = token.assignment_id

            # Add assignment ID to sentry context for debugging
            if hasattr(current_app, 'sentry'):
                current_app.sentry.user_context({
                    'assignment_id': str(g.assignment_id)
                })
        else:
            abort(401)
