from uuid import UUID

from marshmallow import fields, post_dump, post_load, pre_dump, Schema, validate
from sqlalchemy.ext.hybrid import hybrid_property

from stumpf.diva_models import ACTIVITY_TYPES, OCCLUSION_TYPES, \
    SOURCE_TYPES, TRACK_TYPES
from stumpf.models import Fact, FactSchema


# Sub-models stored in the Fact table specific to diva
class TrackDetectionFact(Fact):
    __tablename__ = None
    __mapper_args__ = {
        'polymorphic_identity': 'grozper:track_detection'
    }

    @hybrid_property
    def top(self):
        return self.fact['top']

    @hybrid_property
    def bottom(self):
        return self.fact['bottom']

    @hybrid_property
    def left(self):
        return self.fact['left']

    @hybrid_property
    def right(self):
        return self.fact['right']

    @hybrid_property
    def source(self):
        return self.fact['source']

    @hybrid_property
    def activity(self):
        return self.fact['activity']

    @hybrid_property
    def occlusion(self):
        return self.fact['occlusion']

    @hybrid_property
    def type_(self):
        return self.fact['type']

    @hybrid_property
    def frame(self):
        return self.frame_range.lower


class TrackFact(Fact):
    __tablename__ = None
    __mapper_args__ = {
        'polymorphic_identity': 'grozper:track'
    }

    @hybrid_property
    def type_(self):
        return self.fact['type']

    @hybrid_property
    def detections(self):
        return list(TrackDetectionFact.query.filter(
            TrackDetectionFact.fact['track_id'].astext == str(self.id)
        ))


class ActivityFact(Fact):
    __tablename__ = None
    __mapper_args__ = {
        'polymorphic_identity': 'grozper:activity'
    }

    @hybrid_property
    def type_(self):
        return self.fact['type']

    @hybrid_property
    def tracks(self):
        return list(TrackFact.query.filter(
            TrackFact.fact['activity_id'].astext == str(self.id)
        ))


class GladiatorFact(Fact):
    __tablename__ = None
    __mapper_args__ = {
        'polymorphic_identity': 'gladiator:answer'
    }

    @hybrid_property
    def good(self):
        return self.fact['good']

    @hybrid_property
    def percent_good(self):
        return self.fact['good_count']

    @hybrid_property
    def reasons(self):
        return self.fact['reasons']

    @hybrid_property
    def activity(self):
        return ActivityFact.query.get(self.fact['activity_id'])


# Schemas for the jsonb portion of the fact model.
class DivaFactPropertySchema(Schema):

    # Because we deserialize into a jsonb column, we need uuids to be
    # represented as strings.
    @post_load
    def convert_id_to_string(self, data):
        for key, value in data.items():
            if isinstance(value, UUID):
                data[key] = str(value)


class TrackDetectionFactPropertySchema(DivaFactPropertySchema):
    track_id = fields.UUID(required=True, load_only=True)  # reference to a track fact

    source = fields.Str(required=True, validate=validate.OneOf(SOURCE_TYPES))
    occlussion = fields.Str(missing='none', validate=validate.OneOf(OCCLUSION_TYPES))
    left = fields.Int(required=True)
    right = fields.Int(required=True)
    top = fields.Int(required=True)
    bottom = fields.Int(required=True)

    # The type and activity is duplicated from the TrackFactSchema to help in the very common
    # case of querying detections by activity and track type.
    activity_id = fields.UUID(required=True, load_only=True)  # reference to an activity fact
    type = fields.Str(required=True, validate=validate.OneOf(TRACK_TYPES), load_only=True)


class TrackFactPropertySchema(DivaFactPropertySchema):
    type = fields.Str(required=True, validate=validate.OneOf(TRACK_TYPES))
    activity_id = fields.UUID(required=True, load_only=True)  # reference to another fact
    detections = fields.List(
        fields.Nested('TrackDetectionFactSchema'), dump_only=True)


class ActivityFactPropertySchema(DivaFactPropertySchema):
    type = fields.Str(required=True, validate=validate.OneOf(ACTIVITY_TYPES))
    tracks = fields.List(
        fields.Nested('TrackFactSchema'), dump_only=True)


class GladiatorFactPropertySchema(DivaFactPropertySchema):
    good = fields.Boolean(required=True)
    reasons = fields.List(fields.Str(), missing=list)
    percent_good = fields.Integer(missing=100)
    activity_id = fields.UUID(required=True, load_only=True)  # reference to another fact
    activity = fields.Nested('ActivityFactSchema', dump_only=True)


# Refiner stuff
# Inherit corresponding Grozper Model cuases problem with SQLAlchemy so code is largely duplicated
class RefinerTrackDetectionFact(Fact):
    __tablename__ = None
    __mapper_args__ = {
        'polymorphic_identity': 'refiner:track_detection'
    }

    @hybrid_property
    def top(self):
        return self.fact['top']

    @hybrid_property
    def bottom(self):
        return self.fact['bottom']

    @hybrid_property
    def left(self):
        return self.fact['left']

    @hybrid_property
    def right(self):
        return self.fact['right']

    @hybrid_property
    def source(self):
        return self.fact['source']

    @hybrid_property
    def activity(self):
        return self.fact['activity']

    @hybrid_property
    def occlusion(self):
        return self.fact['occlusion']

    @hybrid_property
    def type_(self):
        return self.fact['type']

    @hybrid_property
    def frame(self):
        return self.frame_range.lower


class RefinerTrackFact(Fact):
    __tablename__ = None
    __mapper_args__ = {
        'polymorphic_identity': 'refiner:track'
    }

    @hybrid_property
    def type_(self):
        return self.fact['type']

    @hybrid_property
    def detections(self):
        return list(RefinerTrackDetectionFact.query.filter(
            RefinerTrackDetectionFact.fact['track_id'].astext == str(self.id)
        ))


class RefinerActivityFact(Fact):
    __tablename__ = None
    __mapper_args__ = {
        'polymorphic_identity': 'refiner:activity'
    }

    @hybrid_property
    def type_(self):
        return self.fact['type']

    @hybrid_property
    def tracks(self):
        return list(RefinerTrackFact.query.filter(
            RefinerTrackFact.fact['activity_id'].astext == str(self.id)
        ))


# Schemas for the full fact models.  These derive from the base FactSchema
# and perform additional mutations to the fact serialization for compatibility
# with existing applications.
class DivaFactSchema(FactSchema):
    @post_dump(pass_original=True)
    def dump_as_fact_property(self, data, original_data):
        factSchema = self.__fact_property_schema__()
        fact = factSchema.dump(original_data.fact)
        fact['id'] = data['id']
        return fact


class TrackDetectionFactSchema(DivaFactSchema):
    __fact_property_schema__ = TrackDetectionFactPropertySchema
    fact = fields.Nested(TrackDetectionFactPropertySchema, load_only=True)

    @post_dump(pass_original=True)
    def set_frame_number(self, data, original_data):
        """Set the frame number from the top-level attribute."""
        data['frame'] = original_data.frame_range.lower
        return data


class TrackFactSchema(DivaFactSchema):
    __fact_property_schema__ = TrackFactPropertySchema
    fact = fields.Nested(TrackFactPropertySchema, load_only=True)

    @pre_dump
    def set_detections_list(self, data):
        data.fact['detections'] = getattr(data, 'detections', [])
        return data


class ActivityFactSchema(DivaFactSchema):
    __fact_property_schema__ = ActivityFactPropertySchema
    fact = fields.Nested(ActivityFactPropertySchema, load_only=True)

    @pre_dump
    def set_detections_list(self, data):
        data.fact['tracks'] = getattr(data, 'tracks', [])
        return data

    @post_dump(pass_original=True)
    def set_compatible_frame_bounds(self, data, original_data):
        data['activity'] = {
            'start_frame': original_data.frame_range.lower,
            'end_frame': original_data.frame_range.upper - 1
        }
        return data


class GladiatorFactSchema(DivaFactSchema):
    __fact_property_schema__ = GladiatorFactPropertySchema
    fact = fields.Nested(GladiatorFactPropertySchema, load_only=True)

    @pre_dump
    def set_activity(self, data):
        data.fact['activity'] = getattr(data, 'activity', None)
        return data


# Refiner Activity Fact capture original grozper activity id
# Adding activty_id to RefinerActivitySchemas
class RefinerActivityFactPropertySchema(ActivityFactPropertySchema):
    activity_id = fields.UUID(required=True, load_only=True)  # reference to another fact
    activity = fields.Nested('ActivityFactSchema', dump_only=True)


class RefinerActivityFactSchema(ActivityFactSchema):
    __fact_property_schema__ = RefinerActivityFactPropertySchema
    fact = fields.Nested(RefinerActivityFactPropertySchema, load_only=True)
