from datetime import datetime
import uuid

from marshmallow import fields, post_dump, validate, validates, ValidationError
from sqlalchemy.dialects import postgresql as pg

from stumpf.models import BaseSchema, db, Range

TRACK_TYPES = ['Person', 'Vehicle', 'Sky', 'Building', 'Shadows', 'Grass', 'Ground', 'Trees',
               'Parking_Lot', 'Street', 'Sidewalk', 'Crosswalk', 'Intersection', 'Stairs',
               'Background', 'Unannotated', 'Door', 'Tree', 'Receptacle', 'Dumpster',
               'Parking_Meter', 'ATM', 'Umbrella', 'Construction_Barrier', 'Bike',
               'Construction_Vehicle', 'Push_Pulled_Object', 'Prop', 'Animal',
               'Articulated_Infrastructure', 'Other', 'Bag', 'Queue']

ACTIVITY_TYPES = ['Sit_Down', 'Stand_Up', 'Carry_in_Hands', 'Carry_on_Back', 'Talk_on_Phone',
                  'Text_On_Phone', 'Read_Document', 'Vehicle_Moving', 'Vehicle_Stopping',
                  'Vehicle_Starting', 'Vehicle_Turning_Left', 'Vehicle_Turning_Right',
                  'Vehicle_UTurn', 'Object_Transfer', 'Pick_Up_Object', 'Set_Down_Object',
                  'Heavy_Carry', 'Hand_Interaction', 'Embrace_Interaction', 'Open_Facility_Door',
                  'Close_Facility_Door', 'Enter_Facility', 'Exit_Facility', 'Open_Vehicle_Door',
                  'Close_Vehicle_Door', 'Enter_Vehicle', 'Exit_Vehicle', 'Open_Trunk',
                  'Close_Trunk', 'Load_Vehicle', 'Unload_Vehicle', 'Vehicle_PicksUp_Person',
                  'Vehicle_DropsOff_Person', 'People_Talking', 'Joining_Queue', 'Riding',
                  'Purchasing', 'Laptop_Interaction', 'Abandon_Package', 'Casing_Facility', 'Theft',
                  'Vehicle_Reversing']

ACTIVITY_TYPE_ENUM = pg.ENUM(*ACTIVITY_TYPES, name='activity_type_enum')
SOURCE_TYPES = ['truth', 'ground-truth', 'linear-interpolation', 'adjudication']
OCCLUSION_TYPES = ['mostly', 'partially', 'none']


class Track(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    video_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('video.id'), nullable=False)
    type_ = db.Column(pg.ENUM(*TRACK_TYPES, name='track_type_enum'), nullable=False)

    video = db.relationship('Video')


class TrackSchema(BaseSchema):
    __model__ = Track

    video_id = fields.UUID(required=True, load_only=True)
    video = fields.Nested('VideoSchema', dump_only=True)
    type_ = fields.Str(required=True, validate=validate.OneOf(TRACK_TYPES))
    detections = \
        fields.List(fields.Nested('DetectionSchema', exclude=('video', 'track', )), dump_only=True)


class Activity(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    video_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('video.id'), nullable=False)
    type_ = db.Column(ACTIVITY_TYPE_ENUM, nullable=False)
    frame_range = db.Column(pg.INT4RANGE, nullable=False)

    video = db.relationship('Video')
    tracks = db.relationship('Track', secondary='activity_track')


class ActivitySchema(BaseSchema):
    __model__ = Activity

    video = fields.Nested('VideoSchema', dump_only=True)
    video_id = fields.UUID(required=True, load_only=True)
    type_ = fields.Str(required=True, validate=validate.OneOf(ACTIVITY_TYPES))
    frame_range = Range(fields.Int(), required=True)
    tracks = fields.List(fields.Nested('TrackSchema', exclude=('video', )), dump_only=True)

    @validates('frame_range')
    def validate_frame_range(self, frame_range):
        # todo: validate against video frame count
        lower, upper = Range.numeric_range_to_list(frame_range)
        if lower >= upper:
            raise ValidationError('Frame range must be at least 1 frame long')


class ArityActivity(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    type_ = db.Column(ACTIVITY_TYPE_ENUM, nullable=False)
    name = db.Column(db.String, nullable=False, unique=True)
    description = db.Column(db.String, nullable=False)


class ActivityGroupSchema(BaseSchema):
    __model__ = Activity

    type_ = fields.Str(required=True, validate=validate.OneOf(ACTIVITY_TYPES))
    frame_range = Range(fields.Int(), required=True)
    tracks = fields.List(fields.Nested('TrackSchema', exclude=('video', )), dump_only=True)

    @validates('frame_range')
    def validate_frame_range(self, frame_range):
        # todo: validate against video frame count
        lower, upper = Range.numeric_range_to_list(frame_range)
        if lower >= upper:
            raise ValidationError('Frame range must be at least 1 frame long')

    @post_dump
    def form_activity_group(self, data):
        activity_group = {
            'type': data['type_'],
            'activity': {
                'start_frame': data['frame_range'][0],
                'end_frame': data['frame_range'][1] - 1
            },
            'tracks': []
        }
        for track in data['tracks']:
            activity_group_track = {
                'type': track['type_'],
                'detections': []
            }
            for detection in track['detections']:
                activity_group_track['detections'].append({
                   'bounds': detection['bounds'],
                   'source': detection['source'],
                   'frame': detection['frame']
                })
            activity_group['tracks'].append(activity_group_track)
        return activity_group


class ActivityTrack(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    activity_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('activity.id'), nullable=False)
    track_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('track.id'), nullable=False)

    activity = db.relationship('Activity')
    track = db.relationship('Track')


class Detection(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    video_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('video.id'), nullable=False)
    track_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('track.id'), nullable=False)
    source = db.Column(pg.ENUM(*SOURCE_TYPES,
                               name='detection_source_enum'), nullable=False)
    occlusion = db.Column(pg.ENUM(*OCCLUSION_TYPES, name='detection_occlusion_enum'),
                          nullable=False)
    frame = db.Column(db.Integer)
    bounds = db.Column(pg.ARRAY(db.Integer))

    video = db.relationship('Video')
    track = db.relationship('Track', backref=db.backref('detections', lazy=True))


class DetectionSchema(BaseSchema):
    __model__ = Detection

    video = fields.Nested('VideoSchema', dump_only=True)
    video_id = fields.UUID(required=True, load_only=True)
    track = fields.Nested('TrackSchema', dump_only=True)
    track_id = fields.UUID(required=True, load_only=True)
    source = fields.Str(required=True, validate=validate.OneOf(SOURCE_TYPES))
    occlusion = fields.Str(required=True, validate=validate.OneOf(OCCLUSION_TYPES))
    frame = fields.Int(required=True)
    bounds = fields.List(fields.Int(), required=True, validate=validate.Length(min=4, max=4))


# TODO: something like this should go here, but I'm too dumb to figure it out.
# event.listen(Assignment.__table__, 'after_create',
#              DDL("""
#     DROP VIEW IF EXISTS view_gladiator_assignments;
#     DROP FUNCTION IF EXISTS gladiator_score_assignment_answer(assignment_id_ UUID);
#
#     CREATE FUNCTION gladiator_score_assignment_answer(assignment_id_ UUID)
#     RETURNS BOOL AS $$
#     DECLARE
#         activity_id_ UUID;
#         task_id_ UUID;
#         is_good_ BOOL;
#         answer_ BOOL;
#     BEGIN
#     SELECT task_id FROM assignment WHERE id = assignment_id_ INTO task_id_;
#     SELECT (answer->>'good')::BOOL FROM assignment WHERE id = assignment_id_ INTO answer_;
#     SELECT (args->>'activityId')::UUID FROM task WHERE id = task_id_ INTO activity_id_;
#     SELECT fact->>'good' FROM fact
#         WHERE (fact->>'activity_id')::UUID = activity_id_
#         AND schema = 'gladiator:answer' LIMIT 1 INTO is_good_;
#     RETURN is_good_ = answer_;
#     END $$ LANGUAGE plpgsql STABLE;
#
#     CREATE VIEW view_gladiator_assignments AS (
#         SELECT gladiator_score_assignment_answer(id) AS correct, * FROM assignment
#         WHERE task_id IN (
#             SELECT id FROM task
#             WHERE annotation_app_id = 'd4f8d27a-b3b6-44c3-b6ed-1e8910048c2d'
#             AND sandbox = false
#         )
#     );
# """))
