from datetime import datetime
import json
import os
import secrets
import string

from faker import Faker
from faker.providers import BaseProvider

from stumpf.models import AnnotationApp, Assignment, AssignmentStatusChange, Task, User, Video

fake = Faker()
alphabet = string.ascii_uppercase + string.digits


class MTurkProvider(BaseProvider):

    def hit_id(self):
        pass

    def assignment_id(self):
        return ''.join(secrets.choice(alphabet) for i in range(30))

    def sns_message(self, events=None):
        return {
            'Type': 'Notification',
            'MessageId': self.generator.uuid4(),
            'TopicArn': 'arn:aws:sns:us-east-2:066423736832:mturk-notifications',
            'Subject': '18151108',
            'Timestamp': '2018-06-19T12:57:07.894Z',
            'SignatureVersion': '1',
            'Signature': 'fake-signature',
            'SigningCertURL': 'https://sns.us-east-2.amazonaws.com/fake-sign-cert-url',
            'UnsubscribeURL': 'https://sns.us-east-2.amazonaws.com/fake-unsub-url',
            'Message': json.dumps({
                'Events': events if events is not None else []
            })
        }

    def sns_task_event(self, **kwargs):
        event = {
            'HITGroupId': ''.join(secrets.choice(alphabet) for i in range(30)),
            'EventType': self.generator.random_element(
                ['HITCreated', 'HITExpired', 'HITReviewable', 'HITDisposed']),
            'EventTimestamp': datetime.utcnow().isoformat(timespec='seconds') + 'Z',
            'HITId': ''.join(secrets.choice(alphabet) for i in range(30)),
            'HITTypeId': ''.join(secrets.choice(alphabet) for i in range(30)),
        }

        event.update(kwargs)

        return event

    def sns_assignment_event(self, **kwargs):
        event = {
            'HITGroupId': ''.join(secrets.choice(alphabet) for i in range(30)),
            'EventType': self.generator.random_element(
                ['AssignmentAbandoned', 'AssignmentReturned', 'AssignmentAccepted',
                 'AssignmentSubmitted', 'AssignmentApproved', 'AssignmentRejected']),
            'EventTimestamp': datetime.utcnow().isoformat(timespec='seconds') + 'Z',
            'HITId': ''.join(secrets.choice(alphabet) for i in range(30)),
            'AssignmentId': ''.join(secrets.choice(alphabet) for i in range(30)),
            'WorkerId': ''.join(secrets.choice(alphabet) for i in range(14)),
            'HITTypeId': ''.join(secrets.choice(alphabet) for i in range(30)),
            'EventDocId': '065146d07bd633ed9dc745c3ed6894cbf44c6182',
            'SourceAccount': '066423736832',
            'CustomerId': 'A2TCFET4XJ6CM1',
            'EventDocVersion': '2014-08-15'
        }

        event.update(kwargs)

        return event


class StumpfProvider(BaseProvider):
    def stumpf_user(self):
        return User(email=self.generator.email(domain='kitware.com'))

    def stumpf_video(self):
        path = os.path.dirname(self.generator.file_path(depth=5))
        name = '%s_%s' % (self.generator.random_int(max=999999),
                          self.generator.file_name(category='video'))

        frame_rate = self.generator.random_element(['30', '60', '29.97', '23.976', '24'])
        duration = self.generator.random_int(max=999999, min=1) / float(frame_rate)
        times = self.generator.random_element([(datetime(2018, 3, 15, 10, 30, 0),
                                                datetime(2018, 3, 15, 10, 35, 0)),
                                               (datetime(2018, 3, 15, 10, 35, 2),
                                                datetime(2018, 3, 15, 10, 40, 2)),
                                               (datetime(2018, 3, 15, 10, 40, 1),
                                                datetime(2018, 3, 15, 10, 45, 1)),
                                               (datetime(2018, 3, 15, 10, 45, 6),
                                                datetime(2018, 3, 15, 10, 50, 6))])
        return Video(**{
            'path': f's3:/%s/%s' % (path, name),
            'name': name,
            'frame_rate': frame_rate,
            'duration': duration,
            'width': self.generator.random_element([1280,
                                                    1440,
                                                    1920,
                                                    1024,
                                                    3840,
                                                    1600]),
            'height': self.generator.random_element([800,
                                                     900,
                                                     1200,
                                                     576,
                                                     2160]),
            'camera_gtag': self.generator.random_element(['G333', 'G421', 'G508', 'G435']),
            'node': self.generator.random_element(['bus', 'school', 'hospital', 'subway']),
            'notes': self.generator.random_element(['few actors and thread activity',
                                                    'lack of soul, partially occluded',
                                                    'down by the schoolyard',
                                                    'watching the tide roll away']),
            'start_time': times[0],
            'end_time': times[1]
        })

    def stumpf_task(self, video_id, annotation_app_id, user_id):
        return Task(**{
            'video_id': video_id,
            'annotation_app_id': annotation_app_id,
            'user_id': user_id,
            'sandbox': self.generator.boolean(),
            'hit_id': self.generator.uuid4(),
            'type_': self.generator.random_element({
                'HIT': 0.95,
                'KWIT': 0.05
            }),
            'title': f'{self.generator.word()} Video annotation',
            'description': self.generator.paragraph(),
            'keywords': ','.join(self.generator.words(nb=5)),
            'assignment_duration_in_seconds': self.generator.random_int(min=60, max=900),
            'status': self.generator.random_element(['Assignable', 'Unassignable', 'Reviewable',
                                                     'Reviewing', 'Disposed']),
            'review_status': self.generator.random_element(
                ['NotReviewed', 'MarkedForReview', 'ReviewedAppropriate', 'ReviewedInappropriate']),
            'max_assignments': self.generator.random_int(min=1, max=5),
            'reward': self.generator.random_int(min=10, max=200),
            'args': self.generator.pydict(3, False, 'str')
        })

    def stumpf_assignment(self, task_id, app_version=None):
        task = Task.query.get(task_id)
        assignment_kwargs = {}

        if task.type_ == 'HIT':
            assignment_kwargs.update({
                'assignment_id': ''.join(secrets.choice(alphabet) for i in range(30)),
                'worker_id': ''.join(secrets.choice(alphabet) for i in range(14)),
                'app_version': app_version if app_version is not None else self.generator.sha1(),
                'status': 'Accepted'
            })

        assignment = Assignment(task_id=task_id, **assignment_kwargs)
        return assignment

    def stumpf_assignment_status_changes(self, assignment):
            assignment_status = self.generator.random_element(
                ['Abandoned', 'Returned', 'Accepted', 'Submitted', 'Approved', 'Rejected'])
            changes = [(AssignmentStatusChange(occurred=datetime.utcnow(),
                                               assignment_id=assignment.id,
                                               status='Accepted'))]

            if assignment_status == 'Accepted':
                return changes
            elif assignment_status in ('Returned', 'Abandoned'):
                changes.append(AssignmentStatusChange(occurred=datetime.utcnow(),
                                                      assignment_id=assignment.id,
                                                      status=assignment_status))
                return changes
            else:
                changes.append(AssignmentStatusChange(occurred=datetime.utcnow(),
                                                      assignment_id=assignment.id,
                                                      status='Submitted'))

            if assignment_status == 'Submitted':
                return changes
            elif assignment_status in ('Approved', 'Rejected'):
                changes.append(AssignmentStatusChange(occurred=datetime.utcnow(),
                                                      assignment_id=assignment.id,
                                                      status=assignment_status))

            return changes

    def stumpf_annotation_app(self, repo_url=None, hash_=None, name=None, url=None):
        return AnnotationApp(**{
            'repo_url': repo_url if repo_url is not None else self.generator.url(),
            'name': name if name is not None else f'{self.generator.last_name()} App',
            'url': url if url is not None else self.generator.url()
        })


fake.add_provider(StumpfProvider)
fake.add_provider(MTurkProvider)
