from datetime import datetime, timedelta
import secrets
from typing import Optional
import uuid

from flask_dance.consumer.backend.sqla import OAuthConsumerMixin
from flask_sqlalchemy import SQLAlchemy
from marshmallow import fields, post_load, pre_load, Schema, validate, validates, ValidationError
from psycopg2.extras import NumericRange
from sqlalchemy import CheckConstraint, DDL, event, MetaData, or_
from sqlalchemy.dialects import postgresql as pg
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.schema import UniqueConstraint

from stumpf.apispec import spec


# This is to avoid having to manually name all constraints
# See: http://alembic.zzzcomputing.com/en/latest/naming.html
metadata = MetaData(naming_convention={
    'pk': 'pk_%(table_name)s',
    'fk': 'fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s',
    'ix': 'ix_%(table_name)s_%(column_0_name)s',
    'uq': 'uq_%(table_name)s_%(column_0_name)s',
    'ck': 'ck_%(table_name)s_%(constraint_name)s',
})
db = SQLAlchemy(metadata=metadata)
event.listen(db.Model.metadata, 'before_create', DDL('CREATE EXTENSION IF NOT EXISTS "uuid-ossp";'))
event.listen(db.Model.metadata, 'before_create', DDL('CREATE EXTENSION IF NOT EXISTS "tablefunc";'))
event.listen(db.Model.metadata, 'before_create', DDL("""
CREATE OR REPLACE FUNCTION log_insert_update_or_delete() RETURNS TRIGGER AS $BODY$
  DECLARE
  channel TEXT;
  payload JSONB;
  ret RECORD;
  BEGIN

  channel = TG_TABLE_NAME || ':' || TG_OP;

  IF(TG_OP = 'DELETE') THEN
    payload = json_build_object('old', to_jsonb(OLD));
    ret = OLD;
  ELSIF (TG_OP = 'UPDATE') THEN
    payload = json_build_object('old', to_jsonb(OLD), 'new', to_jsonb(NEW));
    ret = NEW;
  ELSIF (TG_OP = 'INSERT') THEN
    payload = json_build_object('new', to_jsonb(NEW));
    ret = NEW;
  END IF;

  INSERT INTO log (id, created, category, message) VALUES
         (uuid_generate_v4(), now(), 'NOTIFY',
          json_build_object('channel', channel, 'payload', payload));

  RETURN ret;

  END;
  $BODY$
  LANGUAGE plpgsql VOLATILE;
"""))


class BaseSchema(Schema):
    __model__ = None

    id = fields.UUID()
    created = fields.DateTime(dump_only=True)

    @post_load
    def make_object(self, data):
        return self.__model__(**data)


class Range(fields.List):
    """
    This serializes a numeric interval as a pair of numbers, where the first
    number is the minimum value *inclusive* and the second number is the
    maximum value *exclusive*.  The serialization assumes that the numeric
    ranges are integers.

    For example, the closed interval [1, 100] serializes as follows:

        NumericRange(1, 100, bounds='[]') -> [1, 101]
    """
    @classmethod
    def numeric_range_to_list(cls, value):
        # handle inclusive/exclusive range values correctly
        lower = value.lower
        if not value.lower_inc:
            lower += 1
        upper = value.upper
        if value.upper_inc:
            upper += 1
        return [lower, upper]

    def _deserialize(self, value, attr, data):
        a, b = super()._deserialize(value, attr, data)
        return NumericRange(a, b - 1, bounds='[]')

    def _serialize(self, value, attr, obj):
        return super()._serialize(
            self.numeric_range_to_list(value), attr, obj)


class User(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    email = db.Column(db.String, unique=True, nullable=False)

    @property
    def is_authenticated(self):
        return True

    @property
    def is_active(self):
        return True

    @property
    def is_anonymous(self):
        return not self.is_authenticated

    def get_id(self):
        return self.id


class UserSchema(BaseSchema):
    __model__ = User

    email = fields.Email(required=True)


class OAuth(OAuthConsumerMixin, db.Model):
    provider_user_id = db.Column(db.String(256), unique=True)
    user_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey(User.id))
    user = db.relationship(User)


def now_plus(**kwargs):
    def f():
        return datetime.utcnow() + timedelta(**kwargs)
    return f


class Token(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    user_id = db.Column(db.ForeignKey(User.id, name='token_user_id_fkey'), nullable=False)
    user = db.relationship(User)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    expires = db.Column(db.DateTime, nullable=False, default=now_plus(days=7))
    token = db.Column(db.String, index=True, unique=True, nullable=False, default=secrets.token_hex)

    @classmethod
    def user_from_token(cls, token: str) -> Optional[User]:
        t = cls.query.filter_by(token=token).filter(cls.expires > datetime.utcnow()).first()

        if t:
            return t.user

    @hybrid_property
    def expired(self):
        return self.expires < datetime.utcnow()


event.listen(Token.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE VIEW view_token AS (
                SELECT id, user_id, created, expires
                FROM token)
             """))


class TokenSchema(BaseSchema):
    __model__ = Token

    user_id = fields.UUID(load_only=True)
    user = fields.Nested('UserSchema', dump_only=True)
    expires = fields.DateTime()
    token = fields.Str(dump_only=True)


class MturkToken(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    assignment_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('assignment.id'), nullable=False)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    expires = db.Column(db.DateTime, nullable=False)
    token = db.Column(db.String, unique=True, nullable=False, default=secrets.token_hex)

    assignment = db.relationship('Assignment', backref=db.backref('mturk_tokens', lazy=True))

    @hybrid_property
    def expired(self):
        return self.expires < datetime.utcnow()


event.listen(MturkToken.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE VIEW view_mturk_token AS (
                SELECT id, assignment_id, created, expires
                FROM mturk_token)
             """))


class MturkTokenSchema(BaseSchema):
    __model__ = MturkToken

    expires = fields.DateTime()
    token = fields.Str()
    assignment = fields.Nested('AssignmentSchema', dump_only=True)


class Video(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    path = db.Column(db.String, unique=True, nullable=False)
    name = db.Column(db.String, unique=True, nullable=False)
    frame_rate = db.Column(db.Float, nullable=False)
    duration = db.Column(db.Float, nullable=False)
    width = db.Column(db.Integer, nullable=False)
    height = db.Column(db.Integer, nullable=False)
    camera_gtag = db.Column(db.String, nullable=False)
    start_time = db.Column(db.DateTime(timezone=True), nullable=False)
    end_time = db.Column(db.DateTime(timezone=True), nullable=False)
    node = db.Column(db.String, nullable=False)
    notes = db.Column(db.Text, nullable=True)

    derived_video_id = db.Column(db.ForeignKey('video.id'), nullable=True)
    derived_video = db.relationship('Video', remote_side=[id],
                                    backref=db.backref('derivations'))
    searchable_fields = {
        'path': path,
        'name': name,
        'camera_gtag': camera_gtag,
        'width': width,
        'height': height,
        'node': node,
        'notes': notes
    }
    sortable_fields = {
        'created': created,
        'path': path,
        'name': name,
        'duration': duration,
        'camera_gtag': camera_gtag,
        'start_time': start_time,
        'end_time': end_time
    }
    default_sort = 'name'
    default_desc = False

    def _full_video_frame_range(self):
        return [0, int(self.frame_rate * self.duration)]

    def _handle_fact_generation_defaults(self, frame_range, provenance):
        if provenance is None:
            provenance = {}

        if frame_range is None:
            frame_range = self._full_video_frame_range()

        return frame_range, provenance

    def _fact_from(self, process, fact, task=None, frame_range=None, provenance=None, schema=None):
        frame_range, provenance = self._handle_fact_generation_defaults(frame_range, provenance)

        fact_schema = FactSchema()
        fact = fact_schema.load(
            dict(fact=fact, video_id=self.id, schema=schema, frame_range=frame_range))

        provenance = Provenance(fact=fact, task=task, meta=provenance,
                                relationship=process)
        return fact, provenance

    def get_fact_from_task(self, fact, task, frame_range=None, provenance=None, schema=None):
        return self._fact_from(
            PROV_TASK_CREATED_FACT,
            fact, task=task,
            frame_range=frame_range,
            provenance=provenance,
            schema=schema)

    def get_fact_from_external_process(self, fact, frame_range=None, provenance=None, schema=None):
        return self._fact_from(
            PROV_EXTERNAL_PROCESS_CREATED_FACT,
            fact, frame_range=frame_range,
            provenance=provenance,
            schema=schema)

    def get_fact_from_batch_process(self, fact, frame_range=None, provenance=None, schema=None):
        return self._fact_from(
            PROV_BATCH_PROCESS_CREATED_FACT,
            fact, frame_range=frame_range,
            provenance=provenance,
            schema=schema)


class VideoSchema(BaseSchema):
    __model__ = Video

    path = fields.Url(required=True)
    name = fields.Str(required=True)
    frame_rate = fields.Float(required=True)
    duration = fields.Float(required=True)
    width = fields.Integer(required=True)
    height = fields.Integer(required=True)
    camera_gtag = fields.Str(required=True)
    start_time = fields.DateTime(required=True)
    end_time = fields.DateTime(required=True)
    node = fields.Str(required=True)
    notes = fields.Str(required=False)

    derived_video_id = fields.UUID(load_only=True)
    derived_video = fields.Nested('VideoSchema', dump_only=True)


spec.components.schema('Video', schema=VideoSchema)
spec.tag({'name': 'Video'})


class AnnotationApp(db.Model):
    __table_args__ = (UniqueConstraint('repo_url', 'name', 'url'), )

    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    repo_url = db.Column(db.String, nullable=False)
    name = db.Column(db.String, nullable=False)
    url = db.Column(db.String, nullable=False)

    sortable_fields = {
        'created': created,
        'name': name
    }
    default_sort = 'created'
    default_desc = True


class AnnotationAppSchema(BaseSchema):
    __model__ = AnnotationApp

    repo_url = fields.Url(required=True)
    url = fields.Url(required=True)
    name = fields.Str(required=True)


spec.components.schema('AnnotationApp', schema=AnnotationAppSchema)
spec.tag({'name': 'AnnotationApp'})


class Task(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    user_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('user.id'), nullable=False)
    user = db.relationship('User', backref=db.backref('tasks', lazy=True))
    video_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('video.id'), nullable=False,
                         index=True)
    video = db.relationship('Video', backref=db.backref('tasks', lazy=True))
    annotation_app_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('annotation_app.id'),
                                  nullable=False)
    annotation_app = db.relationship('AnnotationApp', backref=db.backref('tasks', lazy=True))
    sandbox = db.Column(pg.BOOLEAN, nullable=False, default=False)
    hit_id = db.Column(db.String, nullable=True, unique=True)
    hit_group_id = db.Column(db.String, nullable=True)
    type_ = db.Column(pg.ENUM('HIT', 'KWIT', name='hit_kwit_enum'), nullable=False)
    title = db.Column(db.String, nullable=False)
    description = db.Column(db.String, nullable=True)
    keywords = db.Column(db.String, nullable=True)
    assignment_duration_in_seconds = db.Column(db.Integer, nullable=True)
    status = db.Column(pg.ENUM('Assignable', 'Unassignable', 'Reviewable', 'Reviewing', 'Disposed',
                               name='task_status_enum'), nullable=False, default='Assignable')
    review_status = db.Column(pg.ENUM('NotReviewed', 'MarkedForReview', 'ReviewedAppropriate',
                                      'ReviewedInappropriate', name='task_review_status_enum'),
                              nullable=False, default='NotReviewed')
    max_assignments = db.Column(db.Integer, nullable=False)
    reward = db.Column(db.Integer, nullable=True)  # cents
    args = db.Column(pg.JSONB, nullable=False, default=lambda: {})
    answer = db.Column(pg.JSONB, nullable=True)
    qualification_requirements = db.Column(pg.JSONB, nullable=False, default=lambda: [])
    qualification_for_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('qualification_type.id'),
                                     nullable=True)
    qualification_for = db.relationship('QualificationType', backref=db.backref('tasks', lazy=True))

    searchable_fields = {
        'type': type_,
        'task_name': title,
        'assignments': max_assignments
    }

    sortable_fields = {
        'created': created,
        'title': title,
        'type_': type_
    }
    default_sort = 'created'
    default_desc = True

    @property
    def status_changes(self):
        return TaskStatusChange.query.filter_by(task_id=self.id) \
                                     .order_by('occurred').all()


event.listen(Task.__table__, 'after_create',
             DDL("""
             CREATE TRIGGER log_task_insert_update_or_delete
               AFTER INSERT OR UPDATE OR DELETE ON task
               FOR EACH ROW EXECUTE PROCEDURE log_insert_update_or_delete();
             """))


class TaskSchema(BaseSchema):
    __model__ = Task

    annotation_app = fields.Nested('AnnotationAppSchema', dump_only=True)
    title = fields.Str(required=True)
    type_ = fields.Str(required=True)
    video_id = fields.UUID(required=True, load_only=True)
    video = fields.Nested('VideoSchema', dump_only=True)
    sandbox = fields.Boolean(required=True)
    hit_id = fields.Str()
    hit_group_id = fields.Str()
    title = fields.Str(required=True)
    description = fields.Str(required=True)
    keywords = fields.Str()
    assignment_duration_in_seconds = fields.Int()
    status = fields.Str()
    review_status = fields.Str()
    max_assignments = fields.Int(required=True, validate=validate.Range(min=1, max=10))
    reward = fields.Int(required=True, validate=validate.Range(min=1, max=300))
    args = fields.Dict()
    answer = fields.Dict()
    qualification_for_id = fields.UUID(required=False, load_only=True)
    qualification_for = fields.Nested('QualificationTypeSchema', dump_only=True)
    qualification_requirements = fields.List(fields.Dict())


class TaskStatusChange(db.Model):
    __table_args__ = (UniqueConstraint('task_id', 'status'),)

    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    occurred = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    task_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey(Task.id), nullable=False)
    status = db.Column(pg.ENUM('Assignable', 'Unassignable', 'Reviewable', 'Reviewing', 'Disposed',
                               name='task_status_enum'), nullable=False)
    task = db.relationship('Task')


event.listen(TaskStatusChange.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE FUNCTION update_task_status()
             RETURNS TRIGGER AS
             $BODY$
             BEGIN
                UPDATE task SET status = (
                    SELECT status FROM task_status_change
                    WHERE task_id = NEW.task_id
                    ORDER BY occurred DESC LIMIT 1)
                WHERE id = NEW.task_id;
             RETURN NEW;
             END;
             $BODY$
             LANGUAGE plpgsql VOLATILE;"""))
event.listen(TaskStatusChange.__table__, 'after_create',
             DDL("""
             CREATE TRIGGER update_task_status
             AFTER INSERT OR UPDATE ON task_status_change
             FOR EACH ROW EXECUTE PROCEDURE update_task_status();"""))


class CreateTaskSchema(Schema):
    sandbox = fields.Bool(missing=True)
    title = fields.Str(required=True)
    description = fields.Str(required=True)
    keywords = fields.Str()
    max_assignments = fields.Int(required=True, validate=validate.Range(min=1, max=10))
    args = fields.Dict(keys=fields.Str())


class CreateHitSchema(CreateTaskSchema):
    auto_approval_delay_in_seconds = fields.Int(missing=None)
    assignment_duration_in_seconds = fields.Int(required=True,
                                                validate=validate.Range(min=30, max=31536000))
    lifetime_in_seconds = fields.Int(required=True,
                                     validate=validate.Range(min=30, max=31536000))
    reward = fields.Int(required=True, validate=validate.Range(min=1, max=300))
    qualification_requirements = fields.List(fields.Dict(), missing=[])
    args = fields.Dict(keys=fields.Str())
    qualification_for = fields.UUID(required=False)


class Assignment(db.Model):
    __table_args__ = (UniqueConstraint('task_id', 'assignment_id', 'worker_id'), )

    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    task_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey(Task.id), nullable=False)
    assignment_id = db.Column(db.String, nullable=True)
    worker_id = db.Column(db.String, nullable=True)
    app_version = db.Column(db.String, nullable=True, default=None)
    status = db.Column(pg.ENUM('Abandoned', 'Returned', 'Accepted', 'Submitted', 'Approved',
                               'Rejected', name='assignment_status_enum'), default='Accepted',
                       nullable=False)
    answer = db.Column(pg.JSONB, nullable=True)

    task = db.relationship('Task', backref=db.backref('assignments', lazy=True))

    sortable_fields = {
        'created': created
    }
    default_sort = 'created'
    default_desc = True

    @property
    def status_changes(self):
        return AssignmentStatusChange.query.filter_by(assignment_id=self.id) \
                                           .order_by('occurred').all()


event.listen(Assignment.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE VIEW view_assignment_status_times AS (
                SELECT
                assignment_id, accepted, submitted, returned, abandoned, approved,
                EXTRACT(epoch from (submitted - accepted)) as tt_submitted,
                EXTRACT(epoch from (approved - accepted))  as tt_approved,
                EXTRACT(epoch from (returned - accepted))  as tt_returned,
                EXTRACT(epoch from (abandoned - accepted))  as tt_abandoned
                FROM
                CROSSTAB('
                    SELECT cross_product.assignment_id, cross_product.status, s.occurred
                    FROM (
                    SELECT DISTINCT assignment_id, statuses.status
                    FROM assignment_status_change
                    CROSS JOIN (
                        SELECT DISTINCT status FROM assignment_status_change
                        ) AS statuses
                    ) AS cross_product
                    LEFT OUTER JOIN assignment_status_change AS s
                    ON cross_product.assignment_id = s.assignment_id AND
                       cross_product.status = s.status
                    ORDER BY 1, 2
                ') AS (assignment_id UUID,
                    abandoned     TIMESTAMP,
                    returned      TIMESTAMP,
                    accepted      TIMESTAMP,
                    submitted     TIMESTAMP,
                approved TIMESTAMP));
"""))

event.listen(Assignment.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE VIEW view_assignment_detail AS (
                SELECT vast.assignment_id,
                    vast.accepted,
                    vast.submitted,
                    vast.returned,
                    vast.abandoned,
                    vast.approved,
                    vast.tt_submitted,
                    vast.tt_approved,
                    vast.tt_returned,
                    vast.tt_abandoned,
                    a.created AS assignment_created,
                    a.status AS assignment_status,
                    a.worker_id,
                    a.app_version,
                    aa.id AS app_id,
                    aa.name AS app_name,
                    t.id AS task_id,
                    t.title AS task_title,
                    t.max_assignments,
                    t.status AS task_status,
                    t.reward,
                    t.sandbox,
                    t.assignment_duration_in_seconds,
                    v.id AS video_id,
                    v.path,
                    v.frame_rate,
                    v.duration
                FROM view_assignment_status_times vast,
                    assignment a,
                    annotation_app aa,
                    task t,
                    video v
                WHERE a.task_id = t.id AND
                        v.id = t.video_id AND
                        aa.id = t.annotation_app_id AND
                        vast.assignment_id = a.id);
"""))

event.listen(Assignment.__table__, 'after_create',
             DDL("""
             CREATE TRIGGER log_assignment_insert_update_or_delete
               AFTER INSERT OR UPDATE OR DELETE ON assignment
               FOR EACH ROW EXECUTE PROCEDURE log_insert_update_or_delete();
             """))


class AssignmentStatusChange(db.Model):
    __table_args__ = (UniqueConstraint('assignment_id', 'status'),)

    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    occurred = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    assignment_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey(Assignment.id), nullable=False)
    status = db.Column(pg.ENUM('Abandoned', 'Returned', 'Accepted', 'Submitted', 'Approved',
                               'Rejected', name='assignment_status_enum'), default='Accepted',
                       nullable=False)

    assignment = db.relationship('Assignment')


event.listen(AssignmentStatusChange.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE FUNCTION update_assignment_status()
             RETURNS TRIGGER AS
             $BODY$
             BEGIN
                UPDATE assignment SET status = (
                    SELECT status FROM assignment_status_change
                    WHERE assignment_id = NEW.assignment_id
                    ORDER BY occurred DESC LIMIT 1)
                WHERE id = NEW.assignment_id;
             RETURN NEW;
             END;
             $BODY$
             LANGUAGE plpgsql VOLATILE;"""))
event.listen(AssignmentStatusChange.__table__, 'after_create',
             DDL("""
             CREATE TRIGGER update_assignment_status
             AFTER INSERT OR UPDATE ON assignment_status_change
             FOR EACH ROW EXECUTE PROCEDURE update_assignment_status();"""))


event.listen(AssignmentStatusChange.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE VIEW view_assignment_status_change_detail AS (
                SELECT a.id,
                    a.created AS assignment_created,
                    asc_.status AS assignment_status,
                    a.worker_id,
                    a.app_version,
                    aa.id AS app_id,
                    aa.name AS app_name,
                    t.id AS task_id,
                    t.title AS task_title,
                    t.max_assignments,
                    t.status AS task_status,
                    t.reward,
                    t.sandbox,
                    t.assignment_duration_in_seconds,
                    v.id AS video_id,
                    v.path,
                    v.frame_rate,
                    v.duration
                FROM assignment_status_change asc_,
                    assignment a,
                    annotation_app aa,
                    task t,
                    video v
                WHERE a.task_id = t.id AND
                        v.id = t.video_id AND
                        aa.id = t.annotation_app_id AND
                        asc_.assignment_id = a.id);
"""))


class AssignmentStatusChangeSchema(BaseSchema):
    __model__ = AssignmentStatusChange

    created = fields.DateTime(dump_only=True)
    occurred = fields.DateTime()
    assignment_id = fields.Str()
    status = fields.Str()


class AssignmentSchema(BaseSchema):
    __model__ = Assignment

    task = fields.Nested('TaskSchema', dump_only=True)
    assignment_id = fields.Str()
    worker_id = fields.Str()
    status = fields.Str()
    app_version = fields.Str(required=True)
    status_changes = fields.List(fields.Nested('AssignmentStatusChangeSchema'), dump_only=True)


class QualificationType(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    user_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('user.id'), nullable=False)
    user = db.relationship('User', backref=db.backref('qualification_types', lazy=True))
    mturk_qt_id = db.Column(db.String, nullable=True, unique=True)
    sandbox = db.Column(pg.BOOLEAN, nullable=False, default=False)
    name = db.Column(db.String, nullable=False, unique=True)
    description = db.Column(db.String, nullable=True)
    keywords = db.Column(db.String, nullable=True)
    status = db.Column(pg.ENUM('Active', 'Inactive', name='qualification_type_status_enum'),
                       nullable=False, default='Active')


class QualificationTypeSchema(BaseSchema):
    __model__ = QualificationType

    mturk_qt_id = fields.Str()
    sandbox = fields.Boolean()
    name = fields.Str()
    description = fields.Str()
    keywords = fields.Str()
    status = fields.Str()


class CreateQualificationTypeSchema(Schema):
    sandbox = fields.Boolean(missing=True)
    name = fields.Str(required=True)
    description = fields.Str(required=True)
    keywords = fields.Str()
    status = fields.Str(missing='Active', validate=validate.OneOf(['Active', 'Inactive']))


class Log(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    category = db.Column(db.String, nullable=False, index=True)
    message = db.Column(pg.JSONB, nullable=False, default=lambda: {})


event.listen(Log.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE FUNCTION log_notify()
             RETURNS TRIGGER AS
             $BODY$
             BEGIN
                     IF NOT NEW.message ? 'channel' THEN
                        RAISE EXCEPTION 'NOTIFY messages must contain a "channel" attribute';
                     END IF;
                     IF NOT NEW.message ? 'payload' THEN
                        RAISE EXCEPTION 'NOTIFY messages must contain a "payload" attribute';
                     END IF;

                     PERFORM pg_notify('log:INSERT'::text, NEW.id::text);
                     RETURN NEW;
             END;
             $BODY$
             LANGUAGE plpgsql VOLATILE;"""))
event.listen(Log.__table__, 'after_create',
             DDL("""
             CREATE TRIGGER log_notify
             AFTER INSERT ON log
             FOR EACH ROW WHEN (NEW.category = 'NOTIFY')
             EXECUTE PROCEDURE log_notify();"""))


# This view assumes only 1 event per SNS message, which isn't a guarantee - though
# it holds true for all messages received to this point.
event.listen(Log.__table__, 'after_create',
             DDL("""
             CREATE OR REPLACE VIEW view_sns_events AS (
             SELECT id AS log_id,
                    to_timestamp(message->'Message'->'Events'->0->>'EventTimestamp',
                                 'YYYY-MM-DD"T"HH24:MI:SS"Z"')::timestamp at time zone 'UTC'
                                 AS occurred,
                    message->'Message'->'Events'->0->>'HITId' as hit_id,
                    message->'Message'->'Events'->0->>'AssignmentId' as assignment_id,
                    message->'Message'->'Events'->0->>'EventType' as event_type,
                    message->'Message'->'Events'->0->>'WorkerId' as worker_id
             FROM log
             WHERE category = 'raw-sns-message'
             ORDER BY occurred, created)
             """))


class LogSchema(BaseSchema):
    __model__ = Log

    category = fields.Str(required=True)
    message = fields.Dict(required=True)


class Feedback(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    ip = db.Column(db.String, nullable=False)
    user_agent = db.Column(db.String, nullable=False)
    args = db.Column(pg.JSONB, nullable=False, default=lambda: {})


class FeedbackSchema(BaseSchema):
    __model__ = Feedback

    ip = fields.Str(required=True)
    user_agent = fields.Str(required=True)
    args = fields.Dict(required=True)


class Fact(db.Model):

    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    video_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('video.id'), nullable=False,
                         index=True)
    video = db.relationship('Video', backref=db.backref('facts', lazy=True))
    frame_range = db.Column(pg.INT4RANGE, nullable=True)
    fact = db.Column(pg.JSONB, nullable=False)

    # This column differentiates polymorphic types defined elsewhere within the facts
    # table.  A null value here indicates an unknown schema that get resolved to the
    # base class via the `receive_mapper_configured` method below.
    schema = db.Column(db.String, index=True, nullable=True)

    __mapper_args__ = {
        'polymorphic_on': schema
    }


class FactSchema(BaseSchema):
    __model__ = Fact
    video_id = fields.UUID(required=True, load_only=True)
    video = fields.Nested('VideoSchema', dump_only=True)
    frame_range = Range(fields.Int(), required=False)
    fact = fields.Dict(required=True)
    schema = fields.Str(required=False, allow_none=True)

    @validates('frame_range')
    def validate_frame_range(self, frame_range):
        # todo: validate against video frame count
        lower, upper = Range.numeric_range_to_list(frame_range)
        if lower >= upper:
            raise ValidationError('Frame range must be at least 1 frame long')

    @pre_load
    def normalize_frame_range(self, data):
        if isinstance(data['frame_range'], tuple):
            data['frame_range'] = list(data['frame_range'])
        elif isinstance(data['frame_range'], NumericRange):
            data['frame_range'] = Range.numeric_range_to_list(data['frame_range'])
        return data


PROV_TASK_CREATED_FACT = 'Task_Created_Fact'
PROV_EXTERNAL_PROCESS_CREATED_FACT = 'External_Process_Created_Fact'
PROV_BATCH_PROCESS_CREATED_FACT = 'Batch_Process_Created_Fact'
PROV_RELATIONSHIP_TYPES = [
    PROV_TASK_CREATED_FACT,
    PROV_EXTERNAL_PROCESS_CREATED_FACT,
    PROV_BATCH_PROCESS_CREATED_FACT
]


class Provenance(db.Model):
    id = db.Column(pg.UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    created = db.Column(db.DateTime, nullable=False, default=datetime.utcnow)
    fact_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('fact.id'), nullable=False,
                        index=True)
    fact = db.relationship(Fact)
    task_id = db.Column(pg.UUID(as_uuid=True), db.ForeignKey('task.id'), nullable=True,
                        index=True)
    task = db.relationship(Task)
    relationship = db.Column(pg.ENUM(
        *PROV_RELATIONSHIP_TYPES,
        name='provenance_relationship'), nullable=False)
    meta = db.Column(pg.JSONB, nullable=False)

    __table_args__ = (
        CheckConstraint(
            or_(relationship != PROV_TASK_CREATED_FACT, task_id.isnot(None)),
            name='task_id_not_null_if_task_created_Fact'),
    )


class ProvenanceSchema(BaseSchema):
    __model__ = Provenance
    fact_id = fields.UUID(load_only=True, required=True)
    fact = fields.Nested('FactSchema', dump_only=True)
    task_id = fields.UUID(load_only=True, required=False)
    task = fields.Nested('TaskSchema', dump_only=True)
    relationship = fields.Str(required=True, validate=validate.OneOf(PROV_RELATIONSHIP_TYPES))
    meta = fields.Dict(missing=dict)
