from base64 import b64decode
from datetime import datetime
from typing import Dict, Optional

from flask import current_app
from jinja2 import Template
from M2Crypto import X509
import requests
from sqlalchemy.dialects.postgresql import insert

from stumpf.models import Assignment, AssignmentStatusChange, db, Task, TaskStatusChange


external_question_tpl = Template("""<?xml version="1.0"?><ExternalQuestion xmlns="http://mechanicalturk.amazonaws.com/AWSMechanicalTurkDataSchemas/2006-07-14/ExternalQuestion.xsd"><ExternalURL>{{ app_url }}?{{ query_string|escape }}</ExternalURL><FrameHeight>800</FrameHeight></ExternalQuestion>""")  # noqa


def question_xml(**kwargs):
    return external_question_tpl.render(**kwargs)


def generate_message(data: Dict) -> str:
    """
    Takes an SNS data value and generates the signable message.

    This is according to the specification detailed here:
    https://docs.aws.amazon.com/sns/latest/dg/SendMessageToHttp.verify.signature.html
    """
    message_keys = {'Notification': ['Message',
                                     'MessageId',
                                     'Subject',
                                     'Timestamp',
                                     'TopicArn',
                                     'Type'],
                    'SubscriptionConfirmation': ['Message',
                                                 'MessageId',
                                                 'SubscribeURL',
                                                 'Timestamp',
                                                 'Token',
                                                 'TopicArn',
                                                 'Type'],
                    'UnsubscribeConfirmation': ['Message',
                                                'MessageId',
                                                'SubscribeURL',
                                                'Timestamp',
                                                'Token',
                                                'TopicArn',
                                                'Type']}

    keys = message_keys[data['Type']]
    message = ''

    for key in keys:
        message += f'{key}\n{data[key]}\n'

    return message


def verify_signature(data: Dict) -> bool:
    try:
        message = generate_message(data)
        signature = b64decode(data['Signature'])

        cert = X509.load_cert_string(requests.get(data['SigningCertURL']).text)
        pubkey = cert.get_pubkey()
        pubkey.reset_context(md='sha1')
        pubkey.verify_init()
        pubkey.verify_update(message.encode())

        return True if pubkey.verify_final(signature) == 1 else False
    except Exception:
        current_app.logger.warning('Failed when attempting to verify signature')
        return False


def handle_mturk_assignment_event(event: Dict):
    task = Task.query.filter_by(hit_id=event['HITId']).first()

    if not task:
        current_app.logger.warning(f'Received event with unknown task, event: {event}')
        raise Exception('Received event with unknown task')

    event['EventTimestamp'] = datetime.strptime(event['EventTimestamp'], '%Y-%m-%dT%H:%M:%SZ')

    values = {
        'task_id': task.id,
        'assignment_id': event['AssignmentId'],
        'worker_id': event['WorkerId'],
        'status': 'Accepted'
    }

    try:
        """
        Since this endpoint can be processed in parallel, and other endpoints will have the ability
        to create assignments and update their status - this is wrapped in a transaction. Inserting
        a status change will activate a trigger to store the status in the assignment table as well.
        """
        db.session.execute(
            insert(Assignment.__table__).values(**values)
            .on_conflict_do_nothing(constraint='uq_assignment_task_id'))

        assignment = Assignment.query.filter_by(task=task,
                                                assignment_id=event['AssignmentId'],
                                                worker_id=event['WorkerId']).first()
        db.session.execute(
            insert(AssignmentStatusChange.__table__).values(
                assignment_id=assignment.id,
                occurred=event['EventTimestamp'],
                status=event['EventType'].replace('Assignment', ''))
            .on_conflict_do_update(constraint='uq_assignment_status_change_assignment_id',
                                   set_={'occurred': event['EventTimestamp']}))
        db.session.commit()
    except Exception:
        db.session.rollback()
        current_app.logger.warning(f'Failed to create/update assignment, event: {event}')
        raise


def handle_mturk_hit_event(event: Dict):
    task = Task.query.filter_by(hit_id=event['HITId']).first()

    if not task:
        if event['EventType'] == 'HITDisposed':
            # It's possible we deleted the task database side, and received a disposal event long
            # after deletion. If we deleted the task in our database, there's no need to track
            # or warn about this.
            current_app.logger.info('Received HITDisposed event for unknown task.', event)
        else:
            # If a different kind of event comes in for a task that doesn't exist, something
            # went wrong
            current_app.logger.warning('Received event for unknown task', event)
        return
    elif event['EventType'] == 'HITCreated':
        return  # sanity check to ensure the HIT exists

    status_change = TaskStatusChange(task=task)
    if event['EventType'] == 'HITReviewable':
        status_change.status = 'Reviewable'
    elif event['EventType'] == 'HITExpired':
        # From MT blog:
        # "HITs become Unassignable when either the last assignment for that HIT is accepted by
        #  a Worker or when the HIT expires."
        status_change.status = 'Unassignable'
    elif event['EventType'] == 'HITDisposed':
        status_change.status = 'Disposed'
    else:
        current_app.logger.warning('Received unknown event type', event)
        return

    try:
        status_change.occurred = event['EventTimestamp']
        db.session.add(status_change)
        db.session.commit()
    except Exception:
        current_app.logger.warning('Failed to update task', event)
        raise


def create_hit(client, assignment_duration: int, reward: int, title: str,
               description: str, keywords: str, max_assignments: int,
               lifetime_in_seconds: int, question: str,
               auto_approval_delay_in_seconds: Optional[int],
               qualification_requirements: Optional[Dict]) -> Dict:
    hit_type = create_hit_type(client, assignment_duration,
                               reward, title, description, keywords, auto_approval_delay_in_seconds,
                               qualification_requirements)

    return client.create_hit_with_hit_type(HITTypeId=hit_type['HITTypeId'],
                                           MaxAssignments=max_assignments,
                                           LifetimeInSeconds=lifetime_in_seconds,
                                           Question=question)


def create_hit_type(client, assignment_duration: int, reward: int, title: str,
                    description: str, keywords: str,
                    auto_approval_delay_in_seconds: Optional[int],
                    qualification_requirements: Optional[Dict]) -> Dict:
    kwargs = {}

    if auto_approval_delay_in_seconds:
        kwargs['AutoApprovalDelayInSeconds'] = auto_approval_delay_in_seconds

    if qualification_requirements:
        kwargs['QualificationRequirements'] = qualification_requirements

    hit_type = client.create_hit_type(AssignmentDurationInSeconds=assignment_duration,
                                      Reward=str(reward / 100),
                                      Title=title,
                                      Description=description,
                                      Keywords=keywords,
                                      **kwargs)

    client.update_notification_settings(
        HITTypeId=hit_type['HITTypeId'],
        Notification={
            'Destination': current_app.config['MTURK_SNS_ARN'],
            'Transport': 'SNS',
            'Version': '2014-08-15',
            'EventTypes': [
                'AssignmentAccepted',
                'AssignmentAbandoned',
                'AssignmentReturned',
                'AssignmentSubmitted',
                'AssignmentRejected',
                'AssignmentApproved',
                'HITCreated',
                'HITExpired',
                'HITReviewable',
                'HITDisposed']},
        Active=True)

    return hit_type
