import json
import os
import socket

import boto3
from raven import Client

from stumpf.tama import ChibiTama, listen


task_query = """
SELECT answer FROM assignment
    WHERE task_id = '{task_id}'
"""

# at least 3 / 5 or (2 / 3) agreement required
THRESHOLD = 0.66

# aggreement can only be computed if more than this many assignments are complete
MINIMUM_ASSIGNMENTS = 2

ARITY_APP_ID = '2d755948-af16-40bf-846e-0ce885a43b53'


def adjudicate_task(engine, task_id):
    # initialize activity detection count object
    #   activity name ->
    #      [ <# of 0 detections>, <# of 1 detections>, <# of 2+ detections> ]
    activities = {}
    for row in engine.execute(task_query.format(task_id=task_id)):
        if row[0] is None:
            continue
        for activity, value in row[0].items():
            if activity not in activities:
                activities[activity] = [0, 0, 0]
            if value is not None:
                activities[activity][value] += 1

    # adjudicate answers (contains only activities with agreement)
    #   activity -> 0, 1, or 2
    adjudicated = {}
    for activity, counts in activities.items():
        total = sum(counts)
        if not total or len(counts) < MINIMUM_ASSIGNMENTS:
            continue
        for value, count in enumerate(counts):
            if count / total >= THRESHOLD:
                adjudicated[activity] = value
                break

    # post answer back to the task
    engine.execute(f"UPDATE task SET answer = '{json.dumps(adjudicated)}' WHERE id = '{task_id}'")
    return adjudicated


class ArityChibiTama(ChibiTama):
    def __init__(self, *args, **kwargs):
        super(ArityChibiTama, self).__init__(*args, **kwargs)
        Client(os.getenv('SENTRY_DSN'),
               name=f'arity_{socket.gethostname()}',
               processors=('raven.processors.SanitizePasswordsProcessor',
                           'stumpf.sentry.SanitizeProcessor'))
        if 'MTURK_SANDBOX_ENDPOINT_URL' not in os.environ:
            self.log.error(
                'MTURK_SANDBOX_ENDPOINT_URL is not set, '
                'sandbox qualification listeners will fail.'
            )

        if 'MTURK_PRODUCTION_ENDPOINT_URL' not in os.environ:
            self.log.error(
                'MTURK_PRODUCTION_ENDPOINT_URL is not set, '
                'production qualification listeners will fail.'
            )
        try:
            boto3.client('mturk')
        except Exception:
            self.log.exception('AWS credentials are not correctly configured')

    def mturk_client(self, sandbox=True):
        if sandbox:
            endpoint_url = os.environ['MTURK_SANDBOX_ENDPOINT_URL']
        else:
            endpoint_url = os.environ['MTURK_PRODUCTION_ENDPOINT_URL']

        return boto3.client('mturk', endpoint_url=endpoint_url)

    @listen('task:UPDATE')
    def adjudicate_arity_task(self, channel, new=None, old=None, **kwargs):
        if new['annotation_app_id'] != ARITY_APP_ID:
            self.log.debug('Skipping processing of non-arity app')
            return

        if old['status'] != 'Reviewable' and new['status'] == 'Reviewable':
            self.log.info(f'Adjudicating task {new["id"]}')
            answer = adjudicate_task(self.conn, new['id'])
            self.log.debug(f'Task {new["id"]} -> {json.dumps(answer)}')
        else:
            self.log.debug(f'Skipping processing of task in status {new["status"]}')

    @listen('assignment:UPDATE')
    def qualify_worker(self, channel, new=None, old=None, **kwargs):
        if new is None:
            return
        if new.get('answer') is None:
            self.log.debug('Skipping qualification processing for assignment with no answer')
            return
        if new.get('status') != 'Submitted':
            self.log.debug('Skipping qualification processing for non-submitted assignment')
            return

        sandbox, qualification_id = self.conn.execute(
            f'SELECT sandbox, qualification_for_id FROM task WHERE id = \'{new["task_id"]}\''
        ).fetchone()

        qualification = self.fetch_arity_qualification(qualification_id)
        if qualification is None:
            return

        self.log.info(f'Running qualification pipeline on assignment {new["id"]}')
        score = self.score_answer(new)
        params = {
            'WorkerId': new['worker_id'],
            'QualificationTypeId': qualification['mturk_qt_id'],
            'IntegerValue': score,
            'SendNotification': True
        }
        self.log.info(f'Setting worker {new["worker_id"]} score to {score}')
        try:
            self.mturk_client(sandbox).associate_qualification_with_worker(**params)
        except Exception:
            self.log.exception(
                'associate_qualifiction_with_worker failed for worker '
                f'{new["worker_id"]} and qualification {qualification["mturk_qt_id"]}'
            )

    def fetch_arity_qualification(self, qualification_id):
        if qualification_id is None:
            self.log.debug('Task is not a qualification test')
            return
        qualification = self.conn.execute(
            f'SELECT * FROM qualification_type WHERE id = \'{qualification_id}\''
        ).fetchone()
        if qualification is None or 'arity' not in qualification['keywords'].split(' '):
            self.log.debug(f'{qualification_id} is not an arity qualification test')
            return
        return qualification

    def score_answer(self, assignment):
        engine = self.conn
        args, video_id = engine.execute(
            f'SELECT args,video_id FROM task WHERE id = \'{assignment["task_id"]}\''
        ).fetchone()
        activities = json.loads(args['activities']).keys()
        correct = 0
        total = 0
        self.log.debug(f'activities {str(activities)}')
        for activity in activities:
            activity_type = engine.execute(
                f'SELECT type_ FROM arity_activity WHERE name = \'{activity}\''
            ).fetchone()

            if activity_type:
                type_ = activity_type['type_']
                ground_truth_row = engine.execute(
                    'SELECT arity_ground_truth('
                    f"'{video_id}', {args['start']}, {args['end']}, '{type_}')"
                ).fetchone()
                assert ground_truth_row is not None, 'Invalid qualification task'
                ground_truth = ground_truth_row[0]
                total += 1
                self.log.debug(f'ground_truth {ground_truth}')

                answer = assignment['answer'].get(activity)
                self.log.debug(f'answer {answer}')
                if (answer == 0 and ground_truth == '0 times') or \
                        (answer == 1 and ground_truth == '1 time') or \
                        (answer == 2 and ground_truth == '2+ times'):
                    correct += 1
            else:
                self.log.error(f'Invalid activity type {activity}')
        score = int(100 * correct / max(total, 1))
        return score


# For manual execution:
def adjudicate_all_tasks(engine):
    tasks = engine.execute(f"SELECT id FROM task WHERE annotation_app_id = '{ARITY_APP_ID}'")
    count = 0
    for task in tasks:
        count += 1
        adjudicate_task(engine, task['id'])
    return count


if __name__ == '__main__':
    from sqlalchemy import create_engine

    uri = os.environ['STUMPF_POSTGRES_URI']
    engine = create_engine(uri)
    count = adjudicate_all_tasks(engine)
    print(f'Processed {count} tasks')
