from stumpf.app import create_app
from stumpf.diva_fact_schemas import ActivityFact, GladiatorFactPropertySchema
from stumpf.models import Assignment, db, Task
from stumpf.tama import ChibiTama, listen

# all workers must mark the annotation "good" to consider it good
THRESHOLD = 1

# aggreement can only be computed if more than this many assignments are complete
MINIMUM_ASSIGNMENTS = 1

GLADIATOR_APP_ID = 'd4f8d27a-b3b6-44c3-b6ed-1e8910048c2d'

gladiator_schema = GladiatorFactPropertySchema()


def adjudicate_task(task_id):
    app = create_app({'SQLALCHEMY_ECHO': False})
    fact = None

    with app.app_context():
        # get related models and ensure the entities exist
        task = Task.query.get(task_id)
        if task is None:
            raise Exception(f'Task {task_id} not found')

        activity_fact = ActivityFact.query.get(task.args['activityId'])
        if activity_fact is None:
            raise Exception(f'Could not find activity fact associated with task')

        # collect answers from assignments
        good_count = 0
        total_count = 0
        reasons = set()
        for assignment in Assignment.query.filter_by(task_id=task.id):
            if assignment.answer is None:
                continue
            total_count += 1
            if assignment.answer['good']:
                good_count += 1
            reasons.update(assignment.answer.get('badForReasons', []))

        if total_count >= MINIMUM_ASSIGNMENTS:

            # generate a new fact
            fraction_good = good_count / total_count
            fact = gladiator_schema.load({
                'percent_good': int(100 * fraction_good),
                'good': fraction_good >= THRESHOLD,
                'reasons': sorted(reasons),
                'activity_id': str(activity_fact.id)
            })
            video = task.video
            fact_row, provenance_row = video.get_fact_from_task(
                fact, task=task, schema='gladiator:answer',
                frame_range=activity_fact.frame_range)

            db.session.add_all([fact_row, provenance_row])
            db.session.commit()

    return fact


class GladiatorChibiTama(ChibiTama):
    @listen('task:UPDATE')
    def adjudicate_gladiator_task(self, channel, new=None, old=None, **kwargs):
        if new['annotation_app_id'] != GLADIATOR_APP_ID:
            self.log.debug('Skipping processing of non-gladiator app')
            return

        if old['status'] != 'Reviewable' and new['status'] == 'Reviewable':
            self.log.info(f'Adjudicating task {new["id"]}')
            fact = adjudicate_task(new['id'])
            self.log.debug(f'Created fact {fact["id"]}')
        else:
            self.log.debug(f'Skipping processing of task in status {new["status"]}')


if __name__ == '__main__':
    import sys
    for task_id in sys.argv[1:]:
        print(adjudicate_task(task_id))
