from uuid import uuid4

import numpy as np

from stumpf.app import create_app
from stumpf.diva_fact_schemas import ActivityFact, RefinerActivityFactPropertySchema, \
    TrackDetectionFactPropertySchema, TrackFactPropertySchema
from stumpf.models import Assignment, db, Task
from stumpf.tama import ChibiTama, listen

refiner_activity_schema = RefinerActivityFactPropertySchema()
detection_schema = TrackDetectionFactPropertySchema()
track_schema = TrackFactPropertySchema()

REFINER_APP_ID = 'a05be309-1999-45ac-845d-f66eb76ebeba'
REFINER_VERSION = 'refiner_6.0'

PURITY_THRESH_IOU = 0.2
PURITY_THRESH_FRAMES = 0.9
ANN_AGREE_THRESH = 1.5
PURITY_BBOX_IOU = 0.8


def adjudicate_task(task_id, provenance={}):  # noqa
    app = create_app({'SQLALCHEMY_ECHO': False})
    print(task_id)

    with app.app_context():
        # get related models and ensure the entities exist
        task = Task.query.get(task_id)
        if task is None:
            raise Exception(f'Task {task_id} not found')

        video = task.video
        if not video:
            raise Exception(f'Video {task.video_id} not found')

        activity_fact = ActivityFact.query.get(task.args['activityId'])
        if activity_fact is None:
            raise Exception(f'Could not find activity fact associated with task')

        # Step 1: Read in bboxes for track coronation
        assignmentIdDetectionsMap = dict()
        masterTrackIdDetectionsMap = dict()
        assignmentIDs = []
        workerIDs = []

        for assignment in Assignment.query \
            .filter_by(task_id=task.id, app_version=REFINER_VERSION) \
                .filter((Assignment.status == 'Submitted') | (Assignment.status == 'Approved')):
            if assignment.answer is None:
                continue
            assignmentIDs.append(assignment.assignment_id)
            workerIDs.append(assignment.worker_id)
            trackIdDetectionMap = dict()
            activity = assignment.answer['activities'][0]
            tracks = assignment.answer['types']
            detections = assignment.answer['detections']
            for detection in detections:
                if not detection['id1'] in trackIdDetectionMap:
                    trackIdDetectionMap[detection['id1']] = dict()
                trackIdDetectionMap[detection['id1']][detection['frame']] = detection
            assignmentIdDetectionsMap[assignment.assignment_id] = trackIdDetectionMap

        if not len(assignmentIDs):
            return

        track_n_list = list(next(iter(assignmentIdDetectionsMap.values())).keys())

        # Step 2: Determine which tracks should matach
        #    This should remove annotation contributions from lazy or incorrect annotators
        #   -> current method -- track purity
        track_comparison_matrix = np.zeros(
            (len(track_n_list), len(assignmentIDs), len(assignmentIDs)))
        for ii in range(0, len(assignmentIDs)-1):
            for jj in range(ii+1, len(assignmentIDs)):
                for kk in range(0, len(track_n_list)):
                    # 1 or zero based on track purity thresholds, where 1 is track agreement
                    track_comparison_matrix[kk][ii][jj] = track_purity_comparison(
                        assignmentIdDetectionsMap[assignmentIDs[ii]],
                        assignmentIdDetectionsMap[assignmentIDs[jj]],
                        track_n_list[kk])
        # print("track comparison:\n",track_comparison_matrix)

        for ii, track_id in enumerate(track_n_list):
            masterTrackIdDetectionsMap[track_id] = {}
            matching_assignments_pair = track_matching(track_comparison_matrix[ii], assignmentIDs)
            good_annotations = filter_matches(
                matching_assignments_pair, track_id, assignmentIdDetectionsMap)

            if good_annotations:
                # Step 3: Go through each key frame of the track and average good_annotations
                #  -> Remove bad bboxes from contributing to the master bboxes on a per frame basis
                #  -> Weight bbox contribution to average by number of key frames annotated
                #  -> TODO: Only average key frames
                #  -> TODO: weight bbox average such that a tighter bbox gets a heavier weight

                truth_frames = []
                n_key_frames = []
                for good_ann in good_annotations:
                    frames = list(assignmentIdDetectionsMap[good_ann][track_id].keys())
                    frames = [
                        x for x in frames if 'truth' in
                        assignmentIdDetectionsMap[good_ann][track_id][x]['src']]
                    truth_frames.extend(frames)
                    n_key_frames.append(len(frames))
                truth_frames = list(set(truth_frames))
                truth_frames.sort()

                frames = list(assignmentIdDetectionsMap[good_annotations[0]][track_id].keys())
                for frame in frames:
                    bboxes = []
                    for assignment_id in good_annotations:
                        # somehow the frame could mismatch due to some currently
                        # unsolved behavior of Refiner client
                        if frame not in assignmentIdDetectionsMap[assignment_id][track_id]:
                            continue
                        bboxes.append(
                            assignmentIdDetectionsMap[assignment_id][track_id][frame]['bbox'])
                    masterTrackIdDetectionsMap[track_id][frame] = \
                        average_bboxes_weighted_by_keyframes(bboxes, n_key_frames)

            else:
                frames = list(assignmentIdDetectionsMap[assignmentIDs[0]][track_id].keys())
                for frame in frames:
                    masterTrackIdDetectionsMap[track_id][frame] = \
                        assignmentIdDetectionsMap[assignmentIDs[0]][track_id][frame]['bbox']

        activity_fact, activity_provenance = video.get_fact_from_task(
            refiner_activity_schema.load({
                'type': next(iter(activity['act2'])),
                'activity_id': str(activity_fact.id)
            }),
            task,
            frame_range=(activity['timespan'][0]['tsr0'][0],
                         activity['timespan'][0]['tsr0'][1]+1),
            schema='refiner:activity',
            provenance=provenance)
        db.session.add_all([activity_fact, activity_provenance])
        db.session.commit()  # needed to get activity.id for tracks

        for track in tracks:
            actor = [actor for actor in activity['actors'] if actor['id1'] == track['id1']][0]
            track_fact, track_provenance = video.get_fact_from_task(
                track_schema.load({
                    'activity_id': str(activity_fact.id),
                    'type': next(iter(track['cset3']))
                }),
                task,
                frame_range=(actor['timespan'][0]['tsr0'][0], actor['timespan'][0]['tsr0'][1]),
                schema='refiner:track',
                provenance=provenance)
            db.session.add_all([track_fact, track_provenance])
            db.session.commit()  # needed to get track.id for detection

            masterFrameDetectionMap = masterTrackIdDetectionsMap[track['id1']]

            for frame, bbox in masterFrameDetectionMap.items():
                detection, detection_provenance = video.get_fact_from_task(
                    detection_schema.load({
                        'type': track_fact.fact['type'],
                        'activity_id': str(activity_fact.id),
                        'track_id': str(track_fact.id),
                        'source': 'adjudication',
                        'occlusion': 'none',
                        **_convert_bounds_string_to_object(bbox)
                    }),
                    task,
                    frame_range=(frame, frame+1),
                    schema='refiner:track_detection',
                    provenance=provenance)
                db.session.add_all([detection, detection_provenance])
                db.session.commit()

    return activity_fact


def track_purity_comparison(ann1, ann2, track_id):
    """
    This function takes annotated tracks for two annotators and compares them
    for similarity.  The current method for doing this is track purity.
    Track purity measures the percentage of associations with the predominant target
    over the life of the given track.
    INPUTS
    ------
    ann1, ann2 : dict, dict
    track_id : int
    RETURNS
    -------
    agreement : np.array
    """
    if ann1.keys() != ann2.keys():
        print('Track Mismatch')
        return

    agreement = 0
    frames = ann1[track_id].keys()
    # print(frames)
    frame_count = 0
    for frame in frames:
        # somehow the frame could mismatch due to some unsolved behavior of Refiner client
        if frame not in ann2[track_id]:
            continue
        # print(frame in ann2[track_id])
        iou_val = iou(ann1[track_id][frame]['bbox'], ann2[track_id][frame]['bbox'])
        # for each frame, determine if the bboxes overlap and
        # if so increase the purity of that track
        if iou_val > PURITY_THRESH_IOU:
            frame_count += 1
    # if purity of the track is above a threshold, then they agree
    track_purity = frame_count/len(frames)

    if track_purity > PURITY_THRESH_FRAMES:
        agreement = 1

    return agreement


def iou(bbox1, bbox2):
    """
    This function determines intersection over union for two bboxes.
    """
    # determine the (x,y)-coordinates of the intersection rectangle
    xA = max(bbox1[0], bbox2[0])
    yA = max(bbox1[1], bbox2[1])
    xB = max(bbox1[2], bbox2[2])
    yB = max(bbox1[3], bbox2[3])
    # compute the area of intersection rectangle
    interArea = max(0, xB - xA + 1) * max(0, yB - yA + 1)
    # compute the area of both the prediction and ground-truth rectangles
    bbox1Area = (bbox1[2] - bbox1[0] + 1) * (bbox1[3] - bbox1[1] + 1)
    bbox2Area = (bbox2[2] - bbox2[0] + 1) * (bbox2[3] - bbox2[1] + 1)
    # compute the intersection over union
    if float(bbox1Area + bbox2Area - interArea) > 0:
        iou = interArea / float(bbox1Area + bbox2Area - interArea)
    else:
        iou = 0
    return iou


def track_matching(match_matrix, assignmentId):
    """
    This function determines which sets of assignmentId produce agreement on track
    bboxes for activity time interval.
    INPUTS
    ------
    match_matrix : np.array
        zeros and ones, where one indicates track agreement for that pair of assignmentId
    assignmentId : list of str
        list containing identifiers for each assignment
    RETURNS
    -------
    matching_assignments : list of str
        list containing the identifiers for annoators that produced track agreement
    """
    # print("match_matrix", match_matrix)
    matching_tracks = np.nonzero(match_matrix > 0)
    # print('matching_tracks', matching_tracks)
    # print('matching_tracks 2', match_matrix.nonzero())
    matching_assignments_pair = []
    for ii in range(0, len(matching_tracks[0])):
        if assignmentId[matching_tracks[0][ii]] in \
                [x for set in matching_assignments_pair[:] for x in set]:
            for pair in matching_assignments_pair:
                if assignmentId[matching_tracks[0][ii]] in pair \
                        and assignmentId[matching_tracks[1][ii]] not in pair:
                    pair.append(assignmentId[matching_tracks[1][ii]])
        elif assignmentId[matching_tracks[1][ii]] in \
                [x for set in matching_assignments_pair[:] for x in set]:
            for pair in matching_assignments_pair:
                if assignmentId[matching_tracks[1][ii]] in pair \
                        and assignmentId[matching_tracks[0][ii]] not in pair:
                    pair.append(assignmentId[matching_tracks[0][ii]])
        else:
            matching_assignments_pair.append([assignmentId[matching_tracks[0][ii]],
                                              assignmentId[matching_tracks[1][ii]]])
    return matching_assignments_pair


def filter_matches(matching_assignments, track_id, assignmentIdDetectionsMap):
    """
    This function takes the sets of annotators that produce trac agreement and
    determines which particular set should be used for determining refiner result.
    The following parameters are used to determine that the annotations are good:
     - A threshold of annotators must agree on the tracks (ANN_AGREE_THRESH)
     - The % of truth boxes is used as a tie breaker, where more bboxes means
       better annotations.
    INPUTS
    ------
    matching_assignments : list of str
        list containing the identifiers for assignments that produced track agreement
    track_id : int
    assignmentIdDetectionsMap : dict
        dictionary of type geom_kpf output from read_geom_kpf() for each annotator
        geom_kpf[track_id][frame_number]['bbox'] = [TL_x, TL_y, BR_x, BR_y]
        geom_kpf[track_id][frame_number]['src'] contains a string that
            characterizes the annotation on this frame as either
            ground-truth, truth, linear-interpolation
    RETURNS
    -------
    good_annotations : list of str continaing filtered track matches
    """
    if len(matching_assignments) < 1:
        good_annotations = [[]]
    if len(matching_assignments) > 0:
        # threshold of annotators must agree on the assignmentIdDetectionsMap
        good_annotations = [x for x in matching_assignments if len(x) > ANN_AGREE_THRESH-1]
        # number of truth bboxes is tie breaker
        if len(good_annotations) > 1:
            truth_tracks = []
            for pair in good_annotations:
                percent_truths = 0
                for track in pair:
                    n_truths = len(
                        [x for x in assignmentIdDetectionsMap[track][track_id]
                         if 'truth' in assignmentIdDetectionsMap[track][track_id][x]['src']])
                    percent_truths += n_truths/(len(assignmentIdDetectionsMap[track][track_id]))
                truth_tracks.append(percent_truths/len(pair))
            indx = truth_tracks.index(max(truth_tracks))
            good_annotations = [good_annotations[indx]]
    return good_annotations[0]


def average_bboxes_weighted_by_keyframes(bboxes_list, n_key_frames):
    """
    Create average bbox for each key frame of MTurk reported bboxes.intersection weighted by the
    number of key frames produced per annotator
    INPUTS
    ------
    bboxes_list : list of list of ints
        list of bboxes with each bbox of format [TLx, TLy, BRx, BRy]
    n_key_frames : list of ints
        list of number of key frames for each annotator contribution to master
    RETURNS
    -------
    avg_bbox : list of ints
        [TLx, TLy, BRx, BRy]
    """
    n_key_frames = [x-2 for x in n_key_frames]
    avg_bbox = np.zeros(4)
    count_avg_bboxes = 0
    for ii in range(0, len(bboxes_list)):
        bbox = [bboxes_list[ii][0], bboxes_list[ii][1], bboxes_list[ii][2], bboxes_list[ii][3]]
        # BBox Purity: Make sure the bbox overlaps with other bboxes
        bbox_purity = 0
        for bbox2 in bboxes_list:
            if bbox2 != bbox:
                bbox_purity += iou(bbox, bbox2)
        bbox_purity = bbox_purity/(len(bboxes_list)-1)
        # Average BBox: If there is overlap, then create average weighted by number of key frames
        if bbox_purity > PURITY_BBOX_IOU:
            temp_bbox = np.asarray(bbox)
            avg_bbox += temp_bbox*n_key_frames[ii]*100
            count_avg_bboxes += n_key_frames[ii]*100
    # if there is no bbox overlap for this particular frame, then just average all
    if count_avg_bboxes == 0:
        for ii in range(0, len(bboxes_list)):
            temp_bbox = np.asarray(bboxes_list[ii])
            avg_bbox += temp_bbox * n_key_frames[ii]
            count_avg_bboxes += n_key_frames[ii]
    avg_bbox = avg_bbox / count_avg_bboxes
    avg_bbox = [round(float(x), 4) for x in avg_bbox]
    for ii in range(0, len(avg_bbox)):
        if avg_bbox[ii] < 0:
            avg_bbox[ii] == 0.0
    return avg_bbox


def average_bboxes(bboxes_list):
    """
    Create average bbox from list of MTurk reported booxes.intersection
    INPUTS
    ------
    bboxes_list : list of list of ints
        list of bboxes with each bbox of format [TLx, TLy, BRx, BRy]
    RETURNS
    -------
    avg_bbox : list of ints
        [TLx, TLy, BRx, BRy]
    """
    avg_bbox = np.zeros(4)
    count_avg_bboxes = 0
    for bbox in bboxes_list:
        bbox = [bbox[0], bbox[1], bbox[2], bbox[3]]
        # BBox Purity: Make sure the bbox overlaps with other bboxes
        bbox_purity = 0
        for bbox2 in bboxes_list:
            if bbox2 != bbox:
                bbox_purity += iou(bbox, bbox2)
        bbox_purity = bbox_purity/(len(bboxes_list)-1)
        # Average BBox: If there is overlap, then create average
        if bbox_purity > PURITY_BBOX_IOU:
            temp_bbox = np.asarray(bbox)
            avg_bbox += temp_bbox
            count_avg_bboxes += 1
    # if there is no bbox overlap for this particular frame, then just average all
    if count_avg_bboxes == 0:
        for bbox in bboxes_list:
            temp_bbox = np.asarray(bbox)
            avg_bbox += temp_bbox
            count_avg_bboxes += 1
    avg_bbox = avg_bbox/count_avg_bboxes
    avg_bbox = [round(float(x), 4) for x in avg_bbox]
    for ii in range(0, len(avg_bbox)):
        if avg_bbox[ii] < 0:
            avg_bbox[ii] = 0.0
    return avg_bbox


def _convert_bounds_string_to_object(bbox):
    return {
        'left': bbox[0],
        'right': bbox[2],
        'top': bbox[1],
        'bottom': bbox[3]
    }


class RefinerChibiTama(ChibiTama):
    @listen('task:UPDATE')
    def adjudicate_task(self, channel, new=None, old=None, **kwargs):
        if new['annotation_app_id'] != REFINER_APP_ID:
            self.log.debug('Skipping processing of non-refiner app')
            return

        if old['status'] != 'Reviewable' and new['status'] == 'Reviewable':
            self.log.info(f'Adjudicating task {new["id"]}')
            fact = adjudicate_task(new['id'], {
                'chibitama': True
            })
            self.log.debug(f'Created fact {fact["id"]}')
        else:
            self.log.debug(f'Skipping processing of task in status {new["status"]}')


if __name__ == '__main__':
    import sys
    for task_id in sys.argv[1:]:
        print(adjudicate_task(task_id, {
            'batch': str(uuid4())
        }))
