import json
import os
import uuid

from stumpf.tama import ChibiTama, listen

# at least 3 / 5 or (2 / 3) agreement required
THRESHOLD = 0.66
# aggreement can only be computed if more than this many assignments are complete
MINIMUM_ASSIGNMENTS = 2
REID_APP_ID = uuid.UUID('0cd78586-06e9-4ae4-9d46-2e5a98855fe3')
POSSIBLE_VALUES = ['same', 'different', 'unclear']


def adjudicate_task(engine, task_id):
    pairs = {
        # Pairs is a mapping from a set of detections to a string in POSSIBLE_VALUES
        # Example:
        #
        # "4f1f1451-4044-4578-90e0-7aec9737c083.67743332-4044-4578-90e0-7aec9737c083" : 'same',
    }
    task_query = f'SELECT answer FROM assignment WHERE task_id = \'{task_id}\''
    for row in engine.execute(task_query):
        answer_jsonb = row[0]
        # Ignore unanswered tasks
        if answer_jsonb is None:
            continue
        if type(answer_jsonb) is str:
            answer_jsonb = json.loads(answer_jsonb)
        for answer_index, pair in zip(answer_jsonb['answers'], answer_jsonb['pairs']):
            answer = answer_jsonb['answerMap'][answer_index]
            # always sort to make sure reverse comparisons are treated the same.
            pairSet = [pair[0]['id'], pair[1]['id']]
            pairSet.sort()
            pairKey = '.'.join(pairSet)
            if not (pairKey in pairs):
                pairs[pairKey] = {
                    'counts': [0, 0, 0],  # corresponds with POSSIBLE_VALUES
                    'pair': pair,
                }
            value_index = POSSIBLE_VALUES.index(answer['value'])
            pairs[pairKey]['counts'][value_index] += 1

    adjudicated = {val: [] for val in POSSIBLE_VALUES}
    for pair, result in pairs.items():
        counts = result['counts']
        total = sum(counts)
        if not total or total < MINIMUM_ASSIGNMENTS:
            continue
        for index, c in enumerate(counts):
            if c / total >= THRESHOLD:
                val = POSSIBLE_VALUES[index]
                adjudicated[val].append(result['pair'])
                break

    # post answer back to the task
    engine.execute(f"""UPDATE task
                   SET answer= '{json.dumps(adjudicated)}'
                   WHERE id= '{task_id}'""")
    return adjudicated


class ReidChibiTama(ChibiTama):

    @listen('assignment:UPDATE')
    def adjudicate_reid_task(self, channel, new=None, old=None, **kwargs):
        if 'task_id' in new and 'answer' in new:
            task_id = new['task_id']
            answer = new['answer']
            task_query = f'SELECT annotation_app_id FROM task WHERE id = \'{task_id}\' LIMIT 1;'
            app_id = self.conn.execute(task_query).fetchone()[0]

            if app_id != REID_APP_ID:
                self.log.debug('Skipping processing of non-reid app')
                return
            if old['status'] != 'Reviewable' and new['status'] == 'Reviewable':
                self.log.info(f'Adjudicating task {new["id"]}')
                answer = adjudicate_task(self.conn, new['id'])
                self.log.debug(f'Task {new["id"]} -> {json.dumps(answer)}')
            else:
                self.log.debug(f'Skipping processing of task in status {new["status"]}')


# For manual execution:
def adjudicate_all_tasks(engine):
    tasks = engine.execute(f'SELECT id FROM task WHERE annotation_app_id = \'{REID_APP_ID}\'')
    count = 0
    for task in tasks:
        adjudicate_task(engine, task['id'])
        count += 1
    return count


if __name__ == '__main__':
    from sqlalchemy import create_engine

    uri = os.environ['STUMPF_POSTGRES_URI']
    engine = create_engine(uri)
    count = adjudicate_all_tasks(engine)
    print(f'Processed {count} tasks')
