import inspect
import logging
from multiprocessing import Process
import os
import select
import traceback
import uuid

import click
from dotenv import find_dotenv, load_dotenv
import pkg_resources as pr
from psycopg2 import extensions, ProgrammingError
from raven import Client
from sqlalchemy import create_engine, text

load_dotenv(find_dotenv(usecwd=True))

LOG_FORMAT_STRING = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
PIPELINE_ENTRYPOINT = 'stumpf.pipelines'


def listen(*channels):
    """Attach a function to a pg_notify channel.

    This decorator takes one or more arguments and marks a function
    for execution upon receiving a notification on the list of
    channels. It takes a list of channels or may be used multiple
    times. E.g.:

    @listen('foo', 'baz')
    def example(self, channel, msg):
        pass

    @listen('foo')
    @listen('bar')
    @listen('baz')
    def example2(self, channel, msg):
       pass


    The function is not actually associated with with the channel (and
    the channel is not listend too) until the ChibiTama is run (see
    ChibiTama._init_dispatch).
    """
    def _listen(f):
        if not hasattr(f, '_listen_handler'):
            f._listen_handler = []
        f._listen_handler.extend(channels)
        return f
    return _listen


class ChibiTama(Process):
    name = 'ChibiTama'

    def __init__(self, name, postgres_uri, log_level=logging.WARNING, **kwargs):
        super(ChibiTama, self).__init__()
        self.name = name

        self._dispatch_registry = {}
        self._postgres_uri = postgres_uri
        self._engine = None
        self._conn = None

        # Note: sentinel is a property on BaseProcess that
        # accesses _sentinel. In the subprocess _sentinel is not
        # defined and accessing self.sentinel will complete crash
        # the subprocess (wtf python?).
        self._sentinel = None

        self.log = logging.getLogger(self.name)
        self.log.setLevel(log_level)

        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter(LOG_FORMAT_STRING))

        self.log.addHandler(handler)

    def _init_dispatch(self):
        for name, method in inspect.getmembers(self, predicate=inspect.ismethod):
            if hasattr(method, '_listen_handler'):
                # Support multiple channels per method
                for channel in method._listen_handler:
                    self.log.debug(f'Associating method "{method.__qualname__}" with '
                                   f'channel "{channel}"')

                    if channel not in self._dispatch_registry:
                        self._dispatch_registry[channel] = []

                    # Support multiple methods per channel
                    self._dispatch_registry[channel].append(method)

                    self.conn.execute(
                        text(f'LISTEN "log:INSERT"').execution_options(autocommit=True))

    def run(self):
        # http://initd.org/psycopg/docs/advanced.html#asynchronous-notifications
        self._engine = create_engine(self._postgres_uri)
        self.conn = self._engine.connect()
        self.conn.connection.set_isolation_level(extensions.ISOLATION_LEVEL_AUTOCOMMIT)
        self._init_dispatch()

        while True:
            select.select([self.conn.connection], [], [])
            self.conn.connection.poll()
            while self.conn.connection.notifies:
                notify = self.conn.connection.notifies.pop(0)

                logId = None
                try:
                    logId = uuid.UUID(notify.payload)
                except ValueError:
                    self.log.error(f'Payload {notify.payload} was not a valid UUID')
                    continue

                row = None
                try:
                    row = self.conn.execute(f'SELECT message FROM log WHERE id = \'{logId}\';')
                    row = row.fetchone()[0]
                except (ProgrammingError, IndexError) as e:
                    self.log.error(
                        f'processing log table row with id {logId} '
                        f'caused: {e}')
                    continue

                if 'channel' in row and row['channel'] in self._dispatch_registry:
                    try:
                        channel = row['channel']
                        payload = row['payload']
                        self.log.debug(f'Recieved "{channel}" '
                                       f'with payload "{payload}"')
                        # Execute each method associated with this
                        # channel. Note that devs should not rely on
                        # method execution order.
                        for method in self._dispatch_registry[channel]:
                            method(channel, **payload)
                    except Exception:
                        self.log.error(traceback.format_exc())


@click.group()
@click.option('-v', '--verbose', count=True)
@click.pass_context
def tama(ctx, verbose):
    Client(os.getenv('SENTRY_DSN'), environment=os.getenv('ENV', 'development'))
    verbose = min(verbose, 2)

    ctx.obj = {}
    ctx.obj['log'] = logging.getLogger('tama')
    ctx.obj['log'].setLevel([logging.WARNING, logging.INFO, logging.DEBUG][verbose])

    handler = logging.StreamHandler()
    handler.setFormatter(logging.Formatter(LOG_FORMAT_STRING))

    ctx.obj['log'].addHandler(handler)


@tama.command(help='List pipelines that are available to be run')
@click.pass_context
def list(ctx):
    click.echo('Registered Pipelines:')
    for entry_point in pr.iter_entry_points(PIPELINE_ENTRYPOINT):
        ProcessClass = entry_point.load()
        doc_first_line = ProcessClass.__doc__.split('\n')[0] \
            if ProcessClass.__doc__ is not None else 'No Description.'

        click.echo(f'  {entry_point.name} - {doc_first_line}')


@tama.command(help='Run all (or some) pipelines')
@click.option('-d', '--database_uri', envvar='STUMPF_POSTGRES_URI',
              default=False, help='The URI of the postgres database')
@click.argument('pipelines', nargs=-1)
@click.pass_context
def run(ctx, pipelines, database_uri):

    log = ctx.obj['log']

    # Database must be passed in or available at STUMPF_POSTGRES_URI
    # environment variable.
    if not database_uri:
        raise click.BadParameter('STUMPF_POSTGRES_URI environment variable '
                                 'must be set or -p option must be used')

    # Validate Pipeline names passed in. If no pipeline names were
    # passed in assume you want to run all pipelines.
    registered_pipelines = set(
        [ep.name for ep in pr.iter_entry_points(PIPELINE_ENTRYPOINT)])

    pipelines = registered_pipelines if pipelines == () else set(pipelines)

    if len(pipelines - registered_pipelines) == 1:
        raise click.BadParameter(f'{pipelines - registered_pipelines} '
                                 f'is not a valid pipeline name.')

    elif len(pipelines - registered_pipelines) > 1:
        raise click.BadParameter(f'{pipelines - registered_pipelines} '
                                 f'are not valid pipeline names.')

    log.debug(f'Running:  {pipelines}.')
    log.debug(f'Skipping: {registered_pipelines - pipelines}.')

    # Instantiate ChibiTama instances and start their processes.
    procs = []

    for entry_point in pr.iter_entry_points(PIPELINE_ENTRYPOINT):
        if entry_point.name in pipelines:
            ProcessClass = entry_point.load()

            p = ProcessClass(entry_point.name, database_uri, log_level=log.level)
            p.start()

            procs.append(p)

    log.debug(f'All pipelines launched')

    # Block until all pipelines are complete
    for p in procs:
        p.join()
