from datetime import datetime, timedelta

from flask import Blueprint, current_app, g, jsonify, request
from flask_cors import CORS
from marshmallow import fields, post_load, Schema, validates, ValidationError
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.sql import select, update

from stumpf.decorators import public_endpoint
from stumpf.diva_fact_schemas import ActivityFact, ActivityFactSchema
from stumpf.models import (Assignment, AssignmentSchema, AssignmentStatusChange, db, FeedbackSchema,
                           LogSchema, MturkToken, MturkTokenSchema, Task)

activity_schema = ActivityFactSchema()
feedback_schema = FeedbackSchema()
log_schema = LogSchema()
mturk_token_schema = MturkTokenSchema()

mturk_bp = Blueprint('mturk', __name__)
CORS(mturk_bp)


# This is not thread-safe due to the use of self
class MturkAuthSchema(Schema):
    mturk_hit_id = fields.Str(required=True)
    mturk_assignment_id = fields.Str(required=True)
    app_version = fields.Str(required=True)
    mturk_worker_id = fields.Str(required=True)

    @validates('mturk_hit_id')
    def validate_hit_id(self, mturk_hit_id):
        self._task = Task.query.filter_by(hit_id=mturk_hit_id).first()

        if not self._task:
            raise ValidationError('Task does not exist.')

    @post_load
    def make_object(self, data):
        return data, self._task


@mturk_bp.route('/authenticate', methods=['POST'])
@public_endpoint
def authenticate():
    try:
        mturk_auth_schema = MturkAuthSchema()
        auth_data, task = mturk_auth_schema.load(request.json)

        sandbox = ' (sandbox) ' if task.sandbox else ' '

        current_app.logger.debug(f'MTurk{sandbox}authenticate called with auth data: {auth_data}')

        # Attempt to create the assignment, if it was created by SNS
        # already update the app_version (not provided by SNS).
        db.session.execute(
            insert(Assignment.__table__).values(
                task_id=task.id,
                assignment_id=auth_data['mturk_assignment_id'],
                app_version=auth_data['app_version'],
                worker_id=auth_data['mturk_worker_id'],
                status='Accepted')
            .on_conflict_do_update(constraint='uq_assignment_task_id',
                                   set_={'app_version': auth_data['app_version']}))

        # Fetch the assignment row so a status change can be inserted
        assignment_id = db.session.execute(select([Assignment.__table__])
                                           .with_only_columns([Assignment.__table__.c.id])
                                           .where(Assignment.task_id == task.id)
                                           .where(Assignment.assignment_id ==
                                                  auth_data['mturk_assignment_id'])
                                           .where(Assignment.worker_id ==
                                                  auth_data['mturk_worker_id'])
                                           .where(Assignment.app_version ==
                                                  auth_data['app_version'])
                                           .limit(1)).scalar()

        # Insert the status change, ignoring if it's already been logged for this assignment,
        # because that means the time came from SNS (which is the canonical event timestamp)
        db.session.execute(insert(AssignmentStatusChange.__table__).values(
            assignment_id=str(assignment_id), occurred=datetime.utcnow(), status='Accepted')
                           .on_conflict_do_nothing(
                               constraint='uq_assignment_status_change_assignment_id'))

        # Create token for the duration of the HIT assignment value
        expires_at = (datetime.utcnow() +
                      timedelta(seconds=task.assignment_duration_in_seconds))
        token = MturkToken(assignment_id=str(assignment_id), expires=expires_at)

        db.session.add(token)
        db.session.commit()

        return jsonify(mturk_token_schema.dump(token)), 201
    except Exception:
        db.session.rollback()
        raise


# This is not thread-safe due to the use of self
class KwitAuthSchema(Schema):
    task_id = fields.Str(required=True)
    app_version = fields.Str(required=True)

    @validates('task_id')
    def validate_task_id(self, task_id):
        self._task = Task.query.filter_by(id=task_id, status='Assignable').first()

        if not self._task:
            raise ValidationError('Task does not exist or is not assignable.')

    @post_load
    def make_object(self, data):
        return data, self._task


@mturk_bp.route('/authenticate/kwit', methods=['POST'])
@public_endpoint
def authenticate_kwit():
    try:
        kwit_auth_schema = KwitAuthSchema()
        auth_data, task = kwit_auth_schema.load(request.json)

        sandbox = ' (sandbox) ' if task.sandbox else ' '

        current_app.logger.debug(f'KWIT{sandbox}authenticate called with auth data: {auth_data}')

        assignment = Assignment(task=task, app_version=auth_data['app_version'],
                                status='Accepted')
        status_change = AssignmentStatusChange(assignment=assignment, status='Accepted')
        expires_at = datetime.utcnow() + timedelta(days=7)
        token = MturkToken(assignment=assignment, expires=expires_at)

        db.session.add_all([assignment, status_change, token])
        db.session.commit()

        return jsonify(mturk_token_schema.dump(token)), 201
    except Exception:
        db.session.rollback()
        raise


@mturk_bp.route('/activity-group/<uuid:activity_id>', methods=['GET'])
def get_activity_group(activity_id):
    activity = ActivityFact.query.get_or_404(activity_id)
    return jsonify(activity_schema.dump(activity))


@mturk_bp.route('/answer', methods=['POST'])
def submit_answer():
    try:
        schema = AssignmentSchema(only=('id', 'created', 'answer'))
        assignment = schema.load(request.json)
        db.session.execute(update(Assignment.__table__)
                           .where(Assignment.id == g.assignment_id)
                           .values(answer=assignment.answer))
        db.session.commit()

        return jsonify(schema.dump(assignment))
    except Exception:
        db.session.rollback()
        raise


@mturk_bp.route('/log', methods=['POST'])
def submit_log():
    try:
        log = log_schema.load(request.json)
        db.session.add(log)
        db.session.commit()
        return jsonify(log_schema.dump(log)), 201
    except Exception:
        db.session.rollback()
        raise


@mturk_bp.route('/feedback', methods=['POST'])
def submit_feedback():
    try:
        feedback = feedback_schema.load(
            dict({'ip': request.remote_addr,
                  'user_agent': request.headers.get('User-Agent', 'unknown')},
                 **request.json)
        )
        db.session.add(feedback)
        db.session.commit()
        return jsonify(feedback_schema.dump(feedback)), 201
    except Exception:
        db.session.rollback()
        raise
