from urllib.parse import urlencode

import boto3
from flask import Blueprint, current_app, jsonify, request
from flask_login import current_user

from stumpf.models import (AnnotationApp, Assignment, AssignmentSchema, CreateHitSchema,
                           CreateTaskSchema, db, Task, TaskSchema, TaskStatusChange, Video)
from stumpf.mturk_utils import create_hit, question_xml
from stumpf.pagination import paginated_response
from stumpf.sorting import sorted_query


task_bp = Blueprint('task', __name__)

assignment_schema = AssignmentSchema()
task_schema = TaskSchema()
create_hit_schema = CreateHitSchema()
create_task_schema = CreateTaskSchema()


@task_bp.route('/task/<uuid:video_id>/<uuid:app_id>/hit', methods=['POST'])
def create_task_hit(video_id, app_id):
    video = Video.query.get_or_404(video_id)
    annotation_app = AnnotationApp.query.get_or_404(app_id)
    hit_args = create_hit_schema.load(request.json)

    if hit_args['sandbox']:
        mturk = boto3.client('mturk',
                             endpoint_url=current_app.config['MTURK_SANDBOX_ENDPOINT_URL'])
    else:
        mturk = boto3.client('mturk',
                             endpoint_url=current_app.config['MTURK_PRODUCTION_ENDPOINT_URL'])

    try:
        task = Task(video=video, annotation_app=annotation_app, user=current_user, type_='HIT',
                    sandbox=hit_args['sandbox'], title=hit_args['title'],
                    description=hit_args['description'], keywords=hit_args.get('keywords', ''),
                    max_assignments=hit_args['max_assignments'],
                    assignment_duration_in_seconds=hit_args['assignment_duration_in_seconds'],
                    reward=hit_args['reward'], args=hit_args.get('args', {}),
                    qualification_requirements=hit_args['qualification_requirements'],
                    qualification_for_id=hit_args.get('qualification_for'))
        status_change = TaskStatusChange(task=task, status='Assignable')
        db.session.add_all([task, status_change])
        db.session.commit()

        hit = create_hit(mturk, hit_args['assignment_duration_in_seconds'], hit_args['reward'],
                         hit_args['title'], hit_args['description'], hit_args.get('keywords', ''),
                         hit_args['max_assignments'], hit_args['lifetime_in_seconds'],
                         question_xml(app_url=annotation_app.url,
                                      query_string=urlencode({**{'videoId': video.id},
                                                              **hit_args.get('args', {})})),
                         hit_args['auto_approval_delay_in_seconds'],
                         hit_args['qualification_requirements'])

        current_app.logger.info(f'Created HIT {hit["HIT"]["HITId"]} for Task {task.id}')

        task.hit_id = hit['HIT']['HITId']
        task.hit_group_id = hit['HIT']['HITGroupId']
        db.session.add(task)
        db.session.commit()

        return jsonify(task_schema.dump(task)), 201
    except Exception:
        db.session.rollback()
        raise


@task_bp.route('/task/<uuid:video_id>/<uuid:app_id>/kwit', methods=['POST'])
def create_task_kwit(video_id, app_id):
    video = Video.query.get_or_404(video_id)
    annotation_app = AnnotationApp.query.get_or_404(app_id)
    kwit_args = create_task_schema.load(request.json)

    try:
        task = Task(video=video, annotation_app=annotation_app, user=current_user, type_='KWIT',
                    sandbox=kwit_args['sandbox'], title=kwit_args['title'],
                    description=kwit_args['description'], keywords=kwit_args.get('keywords', ''),
                    max_assignments=kwit_args['max_assignments'], args=kwit_args.get('args', {}))
        db.session.add(task)
        db.session.commit()

        return jsonify(task_schema.dump(task)), 201
    except Exception:
        db.session.rollback()
        raise


@task_bp.route('/task', methods=['GET'])
def list_tasks():
    query = sorted_query(Task)
    return paginated_response(query, lambda items: task_schema.dump(items, many=True))


@task_bp.route('/task/<uuid:task_id>/assignments', methods=['GET'])
def list_task_assignments(task_id):
    task = Task.query.get_or_404(task_id)
    query = sorted_query(Assignment, Assignment.query.filter_by(task=task))
    return paginated_response(query, lambda items: assignment_schema.dump(items, many=True))
