from flask import Blueprint, jsonify, request

from stumpf.filtering import filter_multifield_query, filter_query_enum
from stumpf.filtering import filter_query_int, filter_query_string
from stumpf.models import db, Task, TaskSchema, Video, VideoSchema
from stumpf.pagination import paginated_response
from stumpf.sorting import sorted_query


video_bp = Blueprint('video', __name__)
video_schema = VideoSchema()
task_schema = TaskSchema()


@video_bp.route('/video', methods=['POST'])
def create_video():
    """
    ---
    post:
        summary: Create a new video
        description: Create a new video
        tags:
            - Video
        parameters:
            - in: json
              schema: VideoSchema
        responses:
            201:
                schema: VideoSchema
    """
    try:
        video = video_schema.load(request.json)

        if Video.query.filter_by(path=video.path).first():
            return jsonify({
                'error': 'Video path already exists'
            }), 409
        elif Video.query.filter_by(name=video.name).first():
            return jsonify({
                'error': 'Video name already exists'
            }), 409

        db.session.add(video)
        db.session.commit()
        return jsonify(video_schema.dump(video)), 201
    except Exception:
        db.session.rollback()
        raise

    return jsonify(video_schema.dump(video))


@video_bp.route('/video', methods=['GET'])
def list_videos():
    """
    ---
    get:
        summary: List videos
        description: List videos
        tags:
            - Video
        parameters:
          - name: sort
            in: query
            type: string
          - name: desc
            in: query
            type: boolean
        responses:
            200:
                schema:
                    type: array
                    items: VideoSchema
    """
    filteredQuery = filter_multifield_query(Video, Video.searchable_fields.keys(), 'query')
    query = sorted_query(Video, query=filteredQuery)
    return paginated_response(query, lambda items: video_schema.dump(items, many=True))


@video_bp.route('/video/<uuid:video_id>', methods=['GET'])
def get_video(video_id):
    """
    ---
    get:
        summary: Retrieve a single video
        description: Retrieve a single video
        tags:
            - Video
        parameters:
          - name: video_id
            description: UUID of the video
            in: path
            type: string
        responses:
            200:
                schema: VideoSchema
            404:
                description: Video does not exist
    """
    video = Video.query.get_or_404(video_id)
    return jsonify(video_schema.dump(video))


@video_bp.route('/video/<uuid:video_id>/tasks', methods=['GET'])
def list_video_tasks(video_id):
    video = Video.query.get_or_404(video_id)
    query = filter_query_string(Task, 'task_name')
    query = filter_query_int(Task, 'assignments', query)
    query = filter_query_enum(Task, 'type', query)
    query = sorted_query(Task, query.filter_by(video=video))
    return paginated_response(query, lambda items: task_schema.dump(items, many=True))
