import json
import logging
import sys
from uuid import uuid4

import click
import hvac
from stumpf.app import create_app
from stumpf.diva_fact_schemas import ActivityFactPropertySchema, \
                                    TrackDetectionFactPropertySchema, \
                                    TrackFactPropertySchema
from stumpf.models import db, Video

from stumpfwald import cli

activity_schema = ActivityFactPropertySchema()
detection_schema = TrackDetectionFactPropertySchema()
track_schema = TrackFactPropertySchema()


def _convert_bounds_string_to_object(bounds):
    b = bounds.split(' ')
    return {
        'left': int(b[0]),
        'right': int(b[2]),
        'top': int(b[1]),
        'bottom': int(b[3])
    }


@click.group(name='diva', short_help='The DIVA related commands.')
@click.option('--production', is_flag=True, default=False,
              help='Use the production Stumpf database.')
@click.option('--vault-token', type=click.UUID, help='The token to use to authenticate to Vault.')
@click.pass_obj
def diva(ctx, production, vault_token):
    ctx['production'] = production
    ctx['vault_token'] = str(vault_token)


@diva.command(name='import-grozper',
              help='Import activities, tracks, and detections from Grouper JSON.')
@click.argument('video-id', type=click.UUID)
@click.argument('infiles', nargs=-1, type=click.File(), metavar='FILE')
@click.option('--batch', nargs=1, default=str(uuid4()))
@click.pass_obj
def import_grouper(ctx, video_id, infiles, batch):
    logging.getLogger('flask.app').setLevel(logging.ERROR)
    app_config = {
        'SQLALCHEMY_ECHO': False,
    }

    if ctx['production']:
        client = hvac.Client(url='https://vault.kitware.com', token=ctx['vault_token'])
        db_creds = client.read('secret/data/projects/diva-mturk/stumpf/rds')['data']['data']
        app_config['SQLALCHEMY_DATABASE_URI'] = \
            f'postgresql://{db_creds["username"]}:{db_creds["password"]}@{db_creds["host"]}/{db_creds["dbname"]}'  # noqa

    app = create_app(app_config)

    with app.app_context():
        video = Video.query.get(video_id)
        if not video:
            click.echo(f'Video with ID {video_id} does not exist.', err=True)
            sys.exit(1)

        for infile in infiles:
            data = json.load(infile)
            meta = {
                'batch': batch,
                'stumpfwald': True,
                'file': infile.name
            }

            # build activities
            frame_start = data['detections'][0]['ts0']
            frame_end = data['detections'][1]['ts0']

            activity, activity_provenance = video.get_fact_from_external_process(
                activity_schema.load({
                    'type': data['type']
                }),
                frame_range=(frame_start, frame_end),
                schema='grozper:activity',
                provenance=meta
            )
            db.session.add_all([activity, activity_provenance])
            db.session.commit()  # needed to get activity.id for tracks

            for track_object in data['tracks']:
                track_type = track_object['type']
                track, track_provenance = video.get_fact_from_external_process(
                    track_schema.load({
                        'activity_id': str(activity.id),
                        'type': track_type
                    }),
                    frame_range=(frame_start, frame_end),
                    schema='grozper:track',
                    provenance=meta
                )
                db.session.add_all([track, track_provenance])
                db.session.commit()  # needed to get track.id for detection

                for detection_object in track_object['detections']:
                    detection_frame = detection_object['ts0']
                    detection, detection_provenance = video.get_fact_from_external_process(
                        detection_schema.load({
                            'type': track_type,
                            'activity_id': str(activity.id),
                            'track_id': str(track.id),
                            'source': detection_object['src'],
                            'occlusion': 'none',
                            **_convert_bounds_string_to_object(detection_object['g0'])
                        }),
                        frame_range=(detection_frame, detection_frame + 1),
                        schema='grozper:track_detection',
                        provenance=meta
                    )
                    db.session.add_all([detection, detection_provenance])
                db.session.commit()


cli.add_command(diva)
