from subprocess import check_call
from tempfile import NamedTemporaryFile

from stumpf.diva_fact_schemas import RefinerTrackDetectionFact,\
    TrackDetectionFactSchema
from stumpf.models import Assignment, Provenance, Task

tract_detection_schema = TrackDetectionFactSchema()
box_filter_template = (
    "drawbox=enable='between(n,{frame},{frame})':"
    'x={left}:y={top}:w={width}:h={height}:color={color}'
)


def get_refiner_fact_detections(task_id):
    query = Provenance.query.join(Provenance.fact)
    query = query.filter(Provenance.task_id == task_id)
    query = query.order_by(RefinerTrackDetectionFact.frame_range)
    query = query.with_entities(RefinerTrackDetectionFact)
    return tract_detection_schema.dump(query.all(), many=True)


def get_refiner_assignment(assignment_id):
    assignment = Assignment.query.get(assignment_id)
    if assignment is None:
        raise Exception('Assignment id not found')
    return assignment


def get_refiner_answer_detections(assignment):
    detections = []
    for d in assignment.answer['detections']:
        bbox = d['bbox']
        detections.append({
            'left': bbox[0],
            'top': bbox[1],
            'right': bbox[2],
            'bottom': bbox[3],
            'source': d['src'],
            **d
        })
    return detections


def get_refiner_task(task_id):
    task = Task.query.get(task_id)
    if task is None:
        raise Exception('Task id not found')
    return task


def get_video_path_from_assignment(task_id):
    assignment = Assignment.query.get(task_id)
    return assignment.task.video.path


def assemble_box_filter(detection):
    if detection['source'] == 'ground-truth':
        color = 'black'
    elif detection['source'] == 'adjudication':
        color = 'red'
    else:
        color = 'yellow'

    return box_filter_template.format(
        width=(detection['right'] - detection['left']),
        height=(detection['bottom'] - detection['top']),
        color=color,
        **detection
    )


def assemble_clip_filter(detections):
    filters = [assemble_box_filter(detection) for detection in detections]
    start = detections[0]['frame']
    end = detections[-1]['frame']
    filters.append(f"select='between(n,{start},{end})',setpts=PTS-STARTPTS")
    return ','.join(filters)


def generate_clip(detections, input, output):
    with NamedTemporaryFile() as f:
        f.write(assemble_clip_filter(detections).encode())
        f.flush()
        args = ['ffmpeg', '-y', '-loglevel', 'warning', '-i', input, '-filter_script:v', f.name]
        if output == '-':
            args.extend(['-f', 'matroska'])
        args.append(output)
        check_call(args)
