import random
from unittest.mock import patch
from stumpf.diva_models import ACTIVITY_TYPES, TRACK_TYPES
from stumpf.models import db
from faker import Faker
from stumpf.models import Assignment

fake = Faker()


def test_mturk_flask_auth(app):
    with app.test_client() as client:
        resp = client.post('/api/mturk/v1/answer', headers={
            'Stumpf-Mturk-Auth-Token': 'foo'
        })
        assert resp.status_code == 401

        # ensure non-routable routes ignore auth and just 404
        resp = client.post('/api/mturk/v1/non-existent-route', headers={
            'Stumpf-Mturk-Auth-Token': 'foo'
        })
        assert resp.status_code == 404


def test_mturk_authenticate_validation(app):
    with app.test_client() as client:
        resp = client.post('/api/mturk/v1/authenticate', json={
            'mturk_hit_id': 'non-existent',
            'mturk_assignment_id': 'foo',
            'mturk_worker_id': 'bar'
        })

        assert resp.status_code == 400


def test_mturk_authenticate(app, task):
    with app.test_client() as client:
        resp = client.post('/api/mturk/v1/authenticate', json={
            'mturk_hit_id': task.hit_id,
            'mturk_assignment_id': 'foo',
            'app_version': fake.sha1(),
            'mturk_worker_id': 'bar'
        })

        assert resp.status_code == 201


def test_mturk_authenticate_transactional(app, task):
    with app.test_client() as client:
        with patch('stumpf.views.mturk.MturkToken', unsafe=True) as mturk_token:
            mturk_token.side_effect = Exception('some problem')

            resp = client.post('/api/mturk/v1/authenticate', json={
                'mturk_hit_id': task.hit_id,
                'mturk_assignment_id': 'foo',
                'app_version': fake.sha1(),
                'mturk_worker_id': 'bar'
            })

    assert resp.status_code == 500

    # ensure no assignments were created.
    # that is, the entire request failed, not just the token creation
    assert len(Assignment.query.all()) == 0


def test_mturk_authenticate_kwit(app, task):
    with app.test_client() as client:
        task.status = 'Unassignable'
        db.session.add(task)
        db.session.commit()

        resp = client.post('/api/mturk/v1/authenticate/kwit', json={
            'task_id': task.id,
            'app_version': fake.sha1()
        })

        assert resp.status_code == 400, resp.json

        task.status = 'Assignable'
        db.session.add(task)
        db.session.commit()

        resp = client.post('/api/mturk/v1/authenticate/kwit', json={
            'task_id': task.id,
            'app_version': fake.sha1()
        })

        assert resp.status_code == 201, resp.json


def test_mturk_answer_submission(app, task):
    with app.test_client() as client:
        auth_resp = client.post('/api/mturk/v1/authenticate', json={
            'mturk_hit_id': task.hit_id,
            'mturk_assignment_id': 'foo',
            'app_version': fake.sha1(),
            'mturk_worker_id': 'bar'
        })
        assert auth_resp.status_code == 201, resp.json

        for _ in range(2):
            resp = client.post('/api/mturk/v1/answer', headers={
                'X-Stumpf-Mturk-Token': auth_resp.json['token']
            }, json={
                'assignment_id': auth_resp.json['assignment']['id'],
                'answer': {
                    'foo': 'bar'
                }
            })

            assert resp.status_code == 200, resp.json


def test_mturk_answer_submission_kwit(app, task):
    with app.test_client() as client:
        task.status = 'Assignable'
        db.session.add(task)
        db.session.commit()

        auth_resp = client.post('/api/mturk/v1/authenticate/kwit', json={
            'task_id': task.id,
            'app_version': fake.sha1()
        })
        assert auth_resp.status_code == 201, auth_resp.json

        resp = client.post('/api/mturk/v1/answer', headers={
            'X-Stumpf-Mturk-Token': auth_resp.json['token']
        }, json={
            'assignment_id': auth_resp.json['assignment']['id'],
            'answer': {
                'foo': 'bar'
            }
        })

        assert resp.status_code == 200, resp.json

        resp = client.post('/api/mturk/v1/answer', headers={
            'X-Stumpf-Mturk-Token': 'bogus-token'
        }, json={
            'assignment_id': auth_resp.json['assignment']['id'],
            'answer': {
                'foo': 'bar'
            }
        })

        assert resp.status_code == 401, resp.json
